import os
import re
import warnings
from pathlib import Path

import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.time import Time
from requests import ConnectionError, ReadTimeout

from stellarphot.io.tess import (
    TOI,
    TessSubmission,
    TessTargetFile,
    tess_photometry_setup,
)

GOOD_HEADER = {
    "date-obs": "2022-06-04T05:44:28.010",
    "filter": "ip",
    "object": "TIC-237205154",
}

GOOD_HEADER_WITH_PLANET = {
    "date-obs": "2022-06-04T05:44:28.010",
    "filter": "ip",
    "object": "TIC-237205154.01",
}

BAD_HEADER = {}


def test_good_header_sucess():
    tsub = TessSubmission.from_header(GOOD_HEADER)
    assert tsub.utc_start == "20220604"
    assert tsub.tic_id == 237205154
    assert tsub.planet_number == 0
    assert tsub.telescope_code == ""


def test_good_header_planet_sucess():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET)
    assert tsub.utc_start == "20220604"
    assert tsub.tic_id == 237205154
    assert tsub.planet_number == 1
    assert tsub.telescope_code == ""


def test_bad_header_fails():
    with pytest.raises(ValueError) as e:
        TessSubmission.from_header(BAD_HEADER)
    match_str = ".*UTC date of first image.*filter/passband.*TIC ID number.*"
    assert re.search(match_str, str(e))


def test_base_name():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET, telescope_code="ABS")
    assert tsub.base_name == "TIC237205154-01_20220604_ABS_ip"


def test_seeing_profile():
    tsub = TessSubmission.from_header(GOOD_HEADER_WITH_PLANET, telescope_code="ABS")
    assert tsub.seeing_profile == "TIC237205154-01_20220604_ABS_ip_seeing-profile.png"


def test_valid_method():
    tsub = TessSubmission.from_header(GOOD_HEADER)

    # Missing telescope code, amopng other things...
    assert not tsub._valid()

    # Fix the code, should still not be valid because of planet number
    tsub.telescope_code = "ABC"
    assert not tsub._valid()

    # Set planet number and should be valid
    tsub.planet_number = 1
    assert tsub._valid()

    # Set invalid TIC number
    tsub.tic_id = 10_000_000_000
    assert not tsub._valid()


@pytest.mark.remote_data
def test_target_file():
    # Getting the target information failed on windows, so the
    # first point of this test is to simply succeed in creating the
    # object

    tic_742648307 = SkyCoord(ra=104.733225, dec=49.968739, unit="degree")
    server_down = False

    try:
        tess_target = TessTargetFile(tic_742648307, magnitude=12, depth=10)
    except (ConnectionError, ReadTimeout):
        server_down = True
        tess_target = None  # Assure tess_target is defined so that we can delete it
    except ValueError:
        # The server is technically back but producing garbage....
        server_down = True
        tess_target = None  # Assure tess_target is defined so that we can delete it
    else:
        # Check that the first thing in the list is the tick object
        check_coords = SkyCoord(
            ra=tess_target.table["RA"][0],
            dec=tess_target.table["Dec"][0],
            unit=("hour", "degree"),
        )
        assert tic_742648307.separation(check_coords).arcsecond < 1
    finally:
        # Make sure that regardless of whether the server is down we delete
        # the temporary file that gets created.

        # On windows the temporary file cannot be deleted because if it is
        # then the file immediately vanished. That means we get a warning
        # about having an open file handle. We can ignore that warning
        # because we don't care about it here.

        # The warning doesn't get generated until the last reference to
        # the object is deleted, so we need to do that here instead of
        # letting it happen at the end of the test.
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="unclosed file", category=ResourceWarning
            )
            try:
                del tess_target
            except UnboundLocalError:
                pass

        if server_down:
            # The server at University of Louisville is down sometimes, so
            # xfail this test when that happens
            pytest.xfail("TESS/gaia Server down")


class TestTOI:
    @pytest.fixture
    def sample_toi(self):
        toi_dict = {
            "tic_id": 236158940,
            "coord": {
                "ra": "313d25m10.33459171s",
                "dec": "34d21m05.92981069s",
                "representation_type": "spherical",
                "frame": "icrs",
            },
            "depth_ppt": 3.31,
            "depth_error_ppt": 0.13131700000000002,
            "duration": "2.885 h",
            "duration_error": "0.286 h",
            "epoch": {
                "jd1": 2459817.0,
                "jd2": 0.28766999999061227,
                "format": "jd",
                "scale": "tdb",
                "precision": 3,
                "in_subfmt": "*",
                "out_subfmt": "*",
            },
            "epoch_error": "0.0018336 d",
            "period": "2.677573 d",
            "period_error": "1.8e-05 d",
            "tess_mag": 11.3283,
            "tess_mag_error": 0.015,
        }
        return TOI.model_validate(toi_dict)

    @pytest.mark.remote_data
    def test_from_tic_id(self, tess_tic_expected_values):
        # Nothing special about the TIC ID chosen here. It is one we happened
        # to be looking at when writing this test.
        tic_id = tess_tic_expected_values["tic_id"]
        toi_info = TOI.from_tic_id(tic_id)
        assert toi_info.tic_id == tic_id

        # Test the coordinate, but not other properties because those may change
        # over time as the planet candidate is better refined. The coordinate
        # should always be the same.
        expected_coord = tess_tic_expected_values["expected_coords"]
        assert toi_info.coord.separation(expected_coord).arcsecond < 1

        # Try round-tripping through json
        # This is a regression test for #427
        json_str = toi_info.model_dump_json()
        new_toi = TOI.model_validate_json(json_str)

        assert toi_info.coord.separation(new_toi.coord).arcsecond < 0.01

    @pytest.mark.parametrize("start_before_midpoint", [True, False])
    def test_transit_time_for_observation(self, sample_toi, start_before_midpoint):
        # For this test we are checking that the correct transit time is identified
        # for a given observation time.
        # For the sake of the test, we are going to use the 124th transit of the TOI
        # as the reference transit.
        reference_midpoint = sample_toi.epoch + 124 * sample_toi.period
        if start_before_midpoint:
            # Start the observation before the midpoint (and before the transit)
            test_time_start = Time(reference_midpoint - 0.8 * sample_toi.duration)
        else:
            # Start the observation after the midpoint
            test_time_start = Time(reference_midpoint + 0.1 * sample_toi.duration)
        obs_times = test_time_start + np.linspace(0, 2) * sample_toi.duration

        assert sample_toi.transit_time_for_observation(obs_times).jd == pytest.approx(
            reference_midpoint.jd
        )


class TestTessPhotometrySetup:
    # This auto-used fixture changes the working directory to the temporary directory
    # and then changes back to the original directory after the test is done.
    @pytest.fixture(autouse=True)
    def change_to_tmp_dir(self, tmp_path):
        original_dir = os.getcwd()
        os.chdir(tmp_path)
        # Yielding here is important. It means that when the test is done, the remainder
        # of the function will be executed. This is important because the test is run in
        # a temporary directory and we want to change back to the original directory
        # when the test is done.
        yield
        os.chdir(original_dir)

    def test_creation_invalid_input_raises_error(self):
        with pytest.raises(
            ValueError, match="Must provide either TIC ID or TOI object"
        ):
            tess_photometry_setup()

    @pytest.mark.remote_data
    @pytest.mark.parametrize("creation_method", ["tic_id", "toi_object"])
    def test_creation(self, tess_tic_expected_values, creation_method):
        # Check that we can create the necessary files for TESS photometry
        # from a TIC ID.
        tic_id = tess_tic_expected_values["tic_id"]
        if creation_method == "tic_id":
            tess_photometry_setup(tic_id=tic_id)
        else:
            toi_info = TOI.from_tic_id(tic_id)
            tess_photometry_setup(TOI_object=toi_info)

        p_info = Path(f"TIC-{tic_id}-info.json")
        assert p_info.exists()

        p_source_list = Path(f"TIC-{tic_id}-source-list-input.ecsv")
        assert p_source_list.exists()

    @pytest.mark.remote_data
    def test_creation_with_overwrite(self, tess_tic_expected_values):
        # Check to see that an error is raised if the files already exist
        # and the overwrite flag is not set.
        tic_id = tess_tic_expected_values["tic_id"]
        tess_photometry_setup(tic_id=tic_id)

        # Try re-running with both files present, where we hit the error from
        # the source list file first
        with pytest.raises(FileExistsError):
            tess_photometry_setup(tic_id=tic_id)

        # Remove the source list table and try again, this time hitting the
        # error from the info file
        Path(f"TIC-{tic_id}-source-list-input.ecsv").unlink()
        with pytest.raises(FileExistsError):
            tess_photometry_setup(tic_id=tic_id)



import warnings

import numpy as np
from astropy import units as u
from astropy.nddata import CCDData, block_reduce
from astropy.nddata.utils import Cutout2D
from astropy.stats import sigma_clipped_stats
from astropy.table import Table
from astropy.utils.exceptions import AstropyUserWarning
from photutils.detection import DAOStarFinder
from photutils.morphology import data_properties
from photutils.profiles import RadialProfile
from photutils.psf import fit_2dgaussian, fit_fwhm

from stellarphot.core import SourceListData
from stellarphot.settings.models import FwhmMethods

__all__ = ["source_detection", "compute_fwhm", "fast_fwhm_from_image"]


def compute_fwhm(
    ccd,
    sources,
    fwhm_estimate=5,
    x_column="xcenter",
    y_column="ycenter",
    fit_method=FwhmMethods.FIT,  # This matches the old default
    sky_per_pix_avg=None,
    sky_per_pix_column=None,
):
    """
    Computes the FWHM in both x and y directions of sources in an image.

    Parameters
    ----------

    ccd : `astropy.nddata.CCDData` or `numpy.ndarray`
        The CCD Image array.

    sources : `astropy.table.Table`
        An astropy table of the positions of sources in the image.

    fwhm_estimate : float, optional (default=5)
        The initial guess for the FWHM of the sources in the image.

    x_column : str, optional (default='xcenter')
        The name of the column in `sources` that contains the x positions
        of the sources.

    y_column : str, optional (default='ycenter')
        The name of the column in `sources` that contains the y positions
        of the sources.

    fit : bool, optional (default=True)
        If ``True``, fit a 2D Gaussian to each source to estimate its FWHM. If
        ``False``, estimate the FWHM of each source by computing the second
        moments of its light distribution using photutils.

    sky_per_pix_avg : float or array-like of float, optional (default=0)
        Sky background to subtract before centroiding. Cannot specify both
        `sky_per_pix_avg` and `sky_per_pix_column`.

    sky_per_pix_column : str, optional
        The name of the column in `sources` that contains the sky
        background values for each source. This is used to subtract the
        sky background from the cutout before computing the FWHM. Cannot
        specify both `sky_per_pix_avg` and `sky_per_pix_column`.

    Returns
    -------

    fwhm_x, fwhm_y : tuple of np.array float
        The FWHM of each source in the x and y directions.

    """
    data = ccd.data if isinstance(ccd, CCDData) else ccd

    if sky_per_pix_avg is not None and sky_per_pix_column is not None:
        raise ValueError(
            "Cannot specify both `sky_per_pix_avg` and `sky_per_pix_column`."
        )

    if sky_per_pix_column is not None:
        if sky_per_pix_column not in sources.colnames:
            raise ValueError(f"Column {sky_per_pix_column} not found in sources table.")
        sky_values = sources[sky_per_pix_column]

    if sky_per_pix_avg is not None:
        # User gave a value for the sky background to subtract
        sky_values = [sky_per_pix_avg] * len(sources)
    elif sky_per_pix_column is None:
        # User didn't give a value for the sky background to subtract
        # so try an of estimate it from the image

        sky_values = [np.nanmedian(ccd.data)] * len(sources)

    fwhm_x = []
    fwhm_y = []
    for source, sky in zip(sources, sky_values, strict=True):
        x = source[x_column]
        y = source[y_column]

        # Cutout2D needs no units on the center position, so remove unit
        # if it is present.
        try:
            x = x.value
            y = y.value
            sky = sky.value
        except AttributeError:
            pass

        # SKY SUBTRACT STUFF!!
        cutout = Cutout2D(data - sky, (x, y), 5 * fwhm_estimate)

        # Mask any NaNs in the data
        nan_mask = np.isnan(cutout.data)
        inp_mask = getattr(ccd, "mask", None)
        if inp_mask is not None:
            inp_mask = inp_mask[cutout.slices_original]
            mask = inp_mask | nan_mask
        else:
            mask = nan_mask

        cutout_xy = cutout.to_cutout_position((x, y))
        match fit_method:
            case FwhmMethods.FIT:
                # Make sure we get an odd fits shape
                fit_shape = int(2 * ((5 * fwhm_estimate) // 2) + 1)

                # fit_fwhm is supposed to handle NaNs automatically but it doesn't
                # as of photutils 2.2.0
                # see https://github.com/astropy/photutils/issues/2029
                # For now replace any NaN with zero and hope for the best.
                cutout.data[nan_mask] = 0
                fit = fit_fwhm(
                    cutout.data,
                    xypos=cutout_xy,
                    fwhm=fwhm_estimate,
                    fit_shape=fit_shape,
                    mask=mask,
                )
                fit = fit[0]

                fwhm_x.append(fit)  # gaussian_sigma_to_fwhm * fit.x_stddev_1)
                fwhm_y.append(fit)  # gaussian_sigma_to_fwhm * fit.y_stddev_1)
                # print('Still fitting!!')
            case FwhmMethods.MOMENTS:
                sc = data_properties(cutout.data)

                fwhm_xm = sc.fwhm.value
                fwhm_ym = fwhm_xm
                fwhm_x.append(fwhm_xm)
                fwhm_y.append(fwhm_ym)
            case FwhmMethods.PROFILE:
                radii = np.arange(int(3 * fwhm_estimate))
                profile = RadialProfile(cutout.data, cutout_xy, radii)
                fwhm = profile.gaussian_fwhm
                fwhm_x.append(fwhm)
                fwhm_y.append(fwhm)
            case _:
                raise ValueError(f"Unknown fit method: {fit_method}")

    return np.array(fwhm_x), np.array(fwhm_y)


def source_detection(
    ccd,
    fwhm=8,
    sigma=3.0,
    iters=5,
    threshold=10.0,
    stddev=None,
    find_fwhm=True,
    sky_per_pix_avg=0,
    padding=0,
    verbose=False,
):
    """
    Returns an SourceListData object containing the position of sources
    within the image identified using `photutils.DAOStarFinder` algorithm.
    It can also compute the FWHM of each source by fitting a 2D Gaussian
    which can be useful for choosing a useful aperture size for photometry.

    Parameters
    ----------

    ccd : `numpy.ndarray` or `astropy.nddata.CCDData`
        This is the 2-D array of the image to be analyzed.  If
        a CCDData object is passed, its data attribute will be used.

    fwhm : float, optional (default=8)
        Initial estimate of full-width half-max of stars in the image,
        in pixels.

    sigma : float, optional. (default=3.0)
        The number of standard deviations to use as the lower and
        upper clipping limit when calculating the image background.

    iters : int, optional (default=5)
        The number of iterations to perform sigma clipping

    threshold : float, optional. (default=10.0)
        The number of standard deviations above which to select sources.

    stddev : float, optional
        If provided, this value will be used as the standard deviation
        of the image.  If not provided, the standard deviation will be
        calculated using `sigma_clipped_stats` on the image.

    find_fwhm : bool, optional (default=True)
        If ``True``, estimate the FWHM of each source by fitting a 2D Gaussian
        to it. If ``False``, the 'fwhm_x', 'fwhm_y', and 'width' columns of
        output table will be filled with NaN values.

    sky_per_pix_avg : float or None, optional (default=None)
        Sky background to subtract before centroiding. If set to ``None``,
        it will be estimated using the mean of the sigma_clipped_stats
        of the image.

    padding : int, optional (default=0)
        Distance from the edge of the image to ignore when searching for
        sources.

    verbose : bool, optional
        If ``True``, print additional information about the source
        detection process.

    Returns
    -------

    sources: `stellardev.SourceListData`
        A table of the positions of sources in the image.  If `find_fwhm` is
        ``True``, includes columns for `fwhm_x`, `fwhm_y, and `width`.  If
        the input CCDData object has WCS information, the columns `ra` and
        `dec` are also populated, otherwise they are set to NaN.
    """
    # if image is a CCDData object, extract the data array
    if not isinstance(ccd, CCDData) and not isinstance(ccd, np.ndarray):
        raise ValueError("ccd must be a numpy array or CCDData object")

    # If user uses units for fwhm or sky_per_pix_avg, convert to value
    if isinstance(sky_per_pix_avg, u.Quantity):
        sky_per_pix_avg = sky_per_pix_avg.value
    if isinstance(fwhm, u.Quantity):
        fwhm = fwhm.value

    if verbose:
        print(
            "source_detection: You may see a warning about invalid values in the "
            "input image.  This is expected if any pixels are saturated and can be "
            "ignored."
        )

    # Get statistics of the input image (and use them to estimate the sky background
    # if not provided).  Using clipped stats should hopefully get rid of any
    # bright stars that might be in the image. Only do this if stddev is not provided.
    if stddev is None:
        mean, median, stddev = sigma_clipped_stats(ccd, sigma=sigma, maxiters=iters)
    else:
        # Set median to None to indicate sigma clipped stats were not
        # calculated.
        median = None

    if sky_per_pix_avg is None:
        if median is not None:
            # We calculated sigma clipped stats, so use them
            sky_per_pix_avg = median
        else:
            # A median is a pretty good estimate of the sky background, so use it
            # for detection. For *photometry* a better estimate is needed, but for
            # detection, the median is good enough and much faster than sigma clipping.
            sky_per_pix_avg = np.nanmedian(ccd.data)
        if verbose:
            print(f"source_detection: sky_per_pix_avg set to {sky_per_pix_avg:.4f}")

    # Identify sources applying DAOStarFinder to a "sky subtracted"
    # image.
    if verbose:
        print(
            f"source_detection: threshold set to {threshold}* standard deviation "
            f"({stddev:.4f})"
        )
        print(f"source_detection: Assuming fwhm of {fwhm} for DAOStarFinder")

    # daofind should be run on background subtracted image
    # (fails, or at least returns garbage, if sky_per_pix_avg is too low)
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * stddev)
    if isinstance(ccd, CCDData):
        sources = daofind(ccd.data - sky_per_pix_avg)
    else:
        sources = daofind(ccd - sky_per_pix_avg)

    # Identify sources near the edge of the image and remove them
    # from the source list.
    padding_smt = ""
    if padding > 0:
        src_cnt0 = len(sources)
        y_lim, x_lim = ccd.shape
        keep = (
            (sources["xcentroid"].value >= padding)
            & (sources["ycentroid"].value >= padding)
            & (sources["xcentroid"].value < x_lim - padding)
            & (sources["ycentroid"].value < y_lim - padding)
        )
        sources = sources[keep]
        padding_smt = f" (after removing {src_cnt0-len(sources)} sources near the edge)"

    src_cnt = len(sources)
    if verbose:
        print(f"source_detection: {src_cnt} sources identified{padding_smt}.")

    # If image as WCS, compute RA and Dec of each source
    try:
        # Retrieve the RA and Dec of each source as SKyCoord objects, then convert to
        # arrays of floats to add to table
        skypos = ccd.wcs.pixel_to_world(sources["xcentroid"], sources["ycentroid"])
        sources["ra"] = skypos.ra.value
        sources["dec"] = skypos.dec.value
    except AttributeError:
        # No WCS, so add empty columns
        sources["ra"] = np.nan * np.ones(src_cnt)
        sources["dec"] = np.nan * np.ones(src_cnt)

    # If requested, compute the FWHM of each source
    if find_fwhm:
        x, y = compute_fwhm(
            ccd,
            sources,
            fwhm_estimate=fwhm,
            x_column="xcentroid",
            y_column="ycentroid",
            sky_per_pix_avg=sky_per_pix_avg,
        )
        sources["fwhm_x"] = x
        sources["fwhm_y"] = y
        sources["width"] = (x + y) / 2  # Average of x and y FWHM

        # Flag bogus fwhm values returned from fitting (no objects
        # have a fwhm less than 1 pixel)
        bad_src = (sources["fwhm_x"] < 1) | (sources["fwhm_y"] < 1)
        sources["fwhm_x"][bad_src] = np.nan
        sources["fwhm_y"][bad_src] = np.nan
        sources["width"][bad_src] = np.nan
    else:  # add empty columns
        sources["fwhm_x"] = np.nan * np.ones(src_cnt)
        sources["fwhm_y"] = np.nan * np.ones(src_cnt)
        sources["width"] = np.nan * np.ones(src_cnt)

    # Convert sources to SourceListData object by adding
    # units to the columns
    units_dict = {
        "id": None,
        "ra": u.deg,
        "dec": u.deg,
        "xcentroid": u.pix,
        "ycentroid": u.pix,
        "fwhm_x": u.pix,
        "fwhm_y": u.pix,
        "width": u.pix,
    }
    sources = Table(data=sources, units=units_dict)
    # Rename columns to match SourceListData
    colnamemap = {"id": "star_id", "xcentroid": "xcenter", "ycentroid": "ycenter"}

    sl_data = SourceListData(input_data=sources, colname_map=colnamemap)
    return sl_data


def fast_fwhm_from_image(
    ccd,
    fwhm_estimate,
    noise=10,
    n_brightest_sources=30,
    max_adu=40000,
    block_size=8,
    min_block_fwhm=1,
    aggregate_by="mean",
):
    """
    Compute the FWHM of a CCD image by block reducing it, running source detection
    on the reduced image, and then computing the FWHM of the detected sources in the
    original image.

    Parameters
    ----------
    ccd : `numpy.ndarray` or `astropy.nddata.CCDData`
        The CCD image array.
    fwhm_estimate : float
        The initial estimate for the FWHM of the sources in the image.
    noise : float, optional
        The estimated noise in the image.
    n_brightest_sources : int, optional
        The number of brightest sources to use for the FWHM estimate.
    max_adu : float, optional
        The maximum ADU value for a pixel in the image. Sources with peak
        fluxes greater than this value will be ignored.
    block_size : int, optional
        The size of the blocks to use for block reduction. The image will be
        divided into blocks of this size, and source detection will be done on the
        reduced image.
    min_block_fwhm : float, optional
        The minimum FWHM to allow after the block reduction. If the estimated FWHM
        is smaller than this value, the block size will be changed.
    aggregate_by : str, optional
        The method to use for aggregating the FWHM estimates. Can be 'mean' or
        'median'. If None, the FWHM estimates will not be aggregated.

    Returns
    -------
    fwhm : float or np.ndarray
        The FWHM of the sources in the image. If `aggregate_by` is None, this will
        be an array of FWHM values for each source. Otherwise, it will be a single
        value.

    Notes
    -----

    The approach here is to detect sources in the image and measure their FWHM. To
    keep this reasonably fast, we block reduce the image and then run source detection
    on the reduced image. We then use the positions of the detected sources to
    estimate the FWHM of the sources in the original image. Only the brightest
    sources are used to estimate the FWHM, and sources that are too bright (i.e.,
    have a peak flux greater than `max_adu`) are ignored.
    """
    # Check whether the reduced fwhm is larger than the minimum and reset block_size
    # if it is too small.
    if fwhm_estimate / block_size < min_block_fwhm:
        block_size = int(fwhm_estimate / min_block_fwhm)

    # Get data and mask from CCDData object
    if isinstance(ccd, CCDData):
        data = ccd.data
        mask = ccd.mask
    else:
        data = ccd
        mask = None

    with warnings.catch_warnings():
        # block_reduce generates some warnings about things like unit, wcs, etc
        # that are set on ccd but not preserved in the reduced image.
        # That is expected, so ignore it.
        warnings.filterwarnings(
            "ignore",
            message="The following attributes were set on the data object",
            category=AstropyUserWarning,
        )
        reduced_data = block_reduce(data, block_size=block_size)

    # Pick a padding that is about 1% of the block reduced image size
    padding = int(0.01 * reduced_data.shape[0])

    block_sources = source_detection(
        reduced_data,
        fwhm=fwhm_estimate / block_size,
        stddev=noise * block_size,  # noise adds in quadrature
        sky_per_pix_avg=None,  # make source_detection do the sky subtraction
        find_fwhm=False,
        threshold=20,
        padding=padding,
    )

    # Reverse sort by peak flux
    block_sources.sort("peak", reverse=True)
    # To estimate whether the sources are too bright, divide the peak
    # flux by the area of the block.  If this is greater than max_adu,
    # then the sources are too bright.
    too_bright = block_sources["peak"] / block_size**2 > max_adu

    # Drop the sources that are too bright
    block_sources = block_sources[~too_bright]

    # Only keep the n_brightest sources (eventually)
    n_brightest = min(len(block_sources), int(n_brightest_sources))
    # Pad n_brightest a bit in case some of these end up being too bright
    fwhm_est_sources = block_sources[: int(1.2 * n_brightest)]

    # Estimate the x and y positions of the sources in the original image
    fwhm_est_sources["xcenter"] = block_size * (fwhm_est_sources["xcenter"].value + 0.5)
    fwhm_est_sources["ycenter"] = block_size * (fwhm_est_sources["ycenter"].value + 0.5)

    # Compute the FWHM of the sources in the original image
    # Make sure fit_shape is an odd number about 5 times the estimated fwhm
    fit_shape = int(2 * ((5 * fwhm_estimate) // 2) + 1)

    # This next bit is a little sneaky. We mask any values in the image larger than
    # max_adu. After we do the PSF fitting, we only keep results in which there were
    # no masked pixels in the fit. As a result, stars that are too bright in the image
    # will be ignored in the event they made it through the block reduction.
    mask = np.zeros_like(ccd.data, dtype=bool) if mask is None else mask
    mask |= data > max_adu

    with warnings.catch_warnings():
        # fit_2dgaussian generates some warnings if some of the fits do not converge.
        # This probably does not actually affect real images, but it affects tests.
        # We suppress the warnings here and check the fit results for flags that
        # indicate a bad fit.
        warnings.filterwarnings(
            "ignore",
            message=r"One or more fit\(s\) may not have converged. Please check the",
            category=AstropyUserWarning,
        )
        # This is faster than calling our own compute_fwhm, so do this.
        fit = fit_2dgaussian(
            data - np.nanmedian(data),
            xypos=list(
                zip(
                    fwhm_est_sources["xcenter"],
                    fwhm_est_sources["ycenter"],
                    strict=True,
                )
            ),
            fwhm=fwhm_estimate,
            fix_fwhm=False,
            fit_shape=fit_shape,
            mask=mask,
        )

    results = fit.results

    # Only keep good fits -- this is where saturated pixels get removed,
    # because they were masked in the input image and flags=1 means there
    # were masked pixels in the fit.
    results = results[results["flags"] == 0]

    # Estimates of the FWHM that are larger than the fit_shape are bad, so drop
    # any of those
    results = results[results["fwhm_fit"] < fit_shape]

    # Only keep the desired number of sources
    fwhm = results["fwhm_fit"][:n_brightest]

    if aggregate_by is not None:
        match aggregate_by:
            case "mean":
                fwhm = np.mean(fwhm)
            case "median":
                fwhm = np.median(fwhm)
            case _:
                raise ValueError(f"Unknown aggregate_by method: {aggregate_by}")

    return fwhm



import numpy as np
from astropy import units as u
from matplotlib import pyplot as plt

__all__ = [
    "plot_many_factors",
    "bin_data",
    "scale_and_shift",
    "plot_transit_lightcurve",
]


def plot_many_factors(photometry, shift, scale, ax=None):
    """Plots many factors of photometry against each other.

    Parameters
    ----------

    photometry : `stellarphot.PhotometryData`
        The photometry table to plot.

    shift : float
        The amount to shift the data by.

    scale : float
        The amount to scale the data by.

    ax : `matplotlib.axes.Axes`, optional
        The axes to plot on.

    Returns
    -------

    None
        Added features to the plot directly.
    """
    airmass = photometry["airmass"] / np.mean(photometry["airmass"])
    x = photometry["xcenter"] / np.mean(photometry["xcenter"])
    y = photometry["ycenter"] / np.mean(photometry["ycenter"])
    comp_counts = photometry["comparison counts"] / np.mean(
        photometry["comparison counts"]
    )
    sky_per_pix = photometry["sky_per_pix_avg"] / np.mean(photometry["sky_per_pix_avg"])
    width = photometry["width"] / np.mean(photometry["width"])

    scale_airmass = scale_and_shift(airmass, scale, 0.75 * shift, pos=False)
    scale_x = scale_and_shift(x, scale, shift, pos=True)
    scale_y = scale_and_shift(y, scale, shift, pos=True)
    scale_sky_pix = scale_and_shift(sky_per_pix, scale, shift, pos=True)
    scale_counts = scale_and_shift(comp_counts, scale, shift, pos=True)
    scale_width = scale_and_shift(width, scale, shift, pos=True)

    x_times = (photometry["bjd"] - 2400000 * u.day).jd

    if ax is None:
        ax = plt.gca()

    ax.plot(
        x_times,
        scale_counts,
        ".",
        c="brown",
        label="tot_C_cnts (arbitrarily scaled and shifted)",
        alpha=0.5,
        ms=4,
    )
    ax.plot(
        x_times,
        scale_airmass,
        "c-",
        label="AIRMASS (arbitrarily scaled and shifted)",
        ms=4,
    )
    ax.plot(
        x_times,
        scale_sky_pix,
        c="gold",
        label="Sky/Pixel_T1 (arbitrarily scaled and shifted)",
        ms=4,
    )
    ax.plot(
        x_times,
        scale_width,
        "-",
        c="gray",
        label="Width_T1 (arbitrarily scaled and shifted)",
        ms=4,
    )
    ax.plot(
        x_times,
        scale_x,
        "-",
        c="pink",
        label="X(FITS)_T1 (arbitrarily scaled and shifted)",
        ms=4,
    )
    ax.plot(
        x_times,
        scale_y,
        "-",
        c="lightblue",
        label="Y(FITS)_T1 (arbitrarily scaled and shifted)",
        ms=4,
    )


def bin_data(data_set, num=3, error_set=None):
    """Bins data into groups of num.

    Parameters
    ----------

    data_set : array-like
        The data to bin.

    num : int, optional
        The number of data points to bin together. Default is 3.

    error_set : array-like
        The error on the data set. Default is None.

    Returns
    -------

    binned_set : array-like
        The binned data set.

    error : array-like, optional
        The error on the binned data set.
    """
    binned_set = []
    error = []
    for i in range(0, len(data_set), num):
        binned_set.append(data_set[i : i + num].mean())
        if error_set is not None:
            error_bin = error_set[i : i + num] ** 2
            error.append(error_bin.sum() / num)
    return np.array(binned_set), np.array(error)


def scale_and_shift(data_set, scale, shift, pos=True):
    """Scales and shifts data set passed in.

    Parameters
    ----------

    data_set : array-like
        The data to scale and shift.

    scale : float
        The amount to scale the data by.

    shift : float
        The amount to shift the data by (in scaled units).

    pos : bool, optional
        Is data displayed in positive or negative direction? Default is True.

    Returns
    -------

    data_set : array-like
        The scaled and shifted data.
    """
    if not pos:
        data_set = 1 - scale * (data_set - data_set.min()) / (
            data_set.max() - data_set.min()
        )
    else:
        data_set = 1 + scale * (data_set - data_set.min()) / (
            data_set.max() - data_set.min()
        )

    data_set += shift

    return data_set


def plot_transit_lightcurve(
    photometry,
    mod,
    tess_info,
    bin_size,
    low=0.82,
    high=1.06,
):
    # These affect spacing of lines on final plot
    scale = 0.15 * (high - low)
    shift = -0.72 * (high - low)
    # (RMS={rel_flux_rms:.5f})
    grid_y_ticks = np.arange(low, high, 0.02)

    date_obs = photometry["date-obs"][0]
    phot_filter = photometry["passband"][0]
    exposure_time = photometry["exposure"][0]

    midpoint = tess_info.transit_time_for_observation(photometry["bjd"])
    start = midpoint - 0.5 * tess_info.duration
    end = midpoint + 0.5 * tess_info.duration

    detrended_by = []
    if not mod.model.airmass_trend.fixed:
        detrended_by.append("Airmass")
    if not mod.model.spp_trend.fixed:
        detrended_by.append("SPP")
    if not mod.model.width_trend.fixed:
        detrended_by.append("Width")

    flux_full_detrend = mod.data_light_curve(detrend_by="all")
    flux_full_detrend_model = mod.model_light_curve(detrend_by="all")
    rel_detrended_flux = flux_full_detrend / np.mean(flux_full_detrend)

    rel_detrended_flux_rms = np.std(rel_detrended_flux)
    rel_model_rms = np.std(flux_full_detrend_model - rel_detrended_flux)

    rel_flux_rms = np.std(mod.data)

    fig, ax = plt.subplots(1, 1, figsize=(8, 11))

    plt.plot(
        (photometry["bjd"] - 2400000 * u.day).jd,
        photometry["normalized_flux"],
        "b.",
        label=f"rel_flux_T1 (RMS={rel_flux_rms:.5f})",
        ms=4,
    )

    plt.plot(
        mod.times,
        flux_full_detrend - 0.04,
        ".",
        c="r",
        ms=4,
        label=f"rel_flux_T1 ({detrended_by})(RMS={rel_detrended_flux_rms:.5f}), "
        f"(bin size={bin_size})",
    )

    plt.plot(
        mod.times,
        flux_full_detrend - 0.08,
        ".",
        c="g",
        ms=4,
        label=f"rel_flux_T1 ({detrended_by} with transit fit)(RMS={rel_model_rms:.5f}),"
        f" (bin size={bin_size})",
    )
    plt.plot(
        mod.times,
        flux_full_detrend_model - 0.08,
        c="g",
        ms=4,
        label=f"rel_flux_T1 Transit Model ([P={mod.model.period.value:.4f}], "
        f"(Rp/R*)^2={(mod.model.rp.value)**2:.4f}, \na/R*={mod.model.a.value:.4f}, "
        f"[Tc={mod.model.t0.value + 2400000:.4f}], "
        f"[u1={mod.model.limb_u1.value:.1f}, u2={mod.model.limb_u2.value:.1f})",
    )

    plot_many_factors(photometry, shift, scale)

    plt.vlines(start.jd - 2400000, low, 1.025, colors="r", linestyle="--", alpha=0.5)
    plt.vlines(end.jd - 2400000, low, 1.025, colors="r", linestyle="--", alpha=0.5)
    plt.text(
        start.jd - 2400000,
        low + 0.0005,
        f"Predicted\nIngress\n{start.jd-2400000-int(start.jd - 2400000):.3f}",
        horizontalalignment="center",
        c="r",
    )
    plt.text(
        end.jd - 2400000,
        low + 0.0005,
        f"Predicted\nEgress\n{end.jd-2400000-int(end.jd - 2400000):.3f}",
        horizontalalignment="center",
        c="r",
    )

    plt.ylim(low, high)
    plt.xlabel("Barycentric Julian Date (TDB)", fontname="Arial")
    plt.ylabel("Relative Flux (normalized)", fontname="Arial")
    plt.title(
        f"TIC {tess_info.tic_id}.01   UT{date_obs}\nPaul P. Feder Observatory 0.4m "
        f"({phot_filter} filter, {exposure_time} exp, "
        f"fap {photometry['aperture'][0].value:.0f}"
        f"-{photometry['annulus_inner'][0].value:.0f}"
        f"-{photometry['annulus_outer'][0].value:.0f})\n",
        fontsize=14,
        fontname="Arial",
    )
    plt.legend(loc="upper center", frameon=False, fontsize=8, bbox_to_anchor=(0.6, 1.0))
    ax.set_yticks(grid_y_ticks)
    plt.grid()

    plt.savefig(
        f"TIC{tess_info.tic_id}-01_{date_obs}_Paul-P-Feder-0.4m_{phot_filter}_lightcurve.png",
        facecolor="w",
    )



from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any

import numpy as np
from astropy.utils.data import get_pkg_data_path
from pydantic import BaseModel, BeforeValidator, Field

__all__ = [
    "MagnitudeSystemNames",
    "MagnitudeSystem",
    "MagnitudeTransformPolynomial",
    "MagnitudeSystemTransform",
    "PanStarrs1ToJohnsonCousins",
]


class MagnitudeSystemNames(StrEnum):
    """Enum for different magnitude systems."""

    AB = "AB"
    Vega = "Vega"
    JC = "JC"
    SDSSDR7 = "SDSSDR7"
    USNO_SDSS = "USNO_SDSS"
    GAIA = "GAIA"
    TESS = "TESS"
    PANSTARRS1 = "PANSTARRS1"
    UGRIZ = "ugriz"


# class MagnitudePassband(BaseModel):
#     name: str
#     # passband_range: tuple[float, float]
#     # Down the road add something like symphot.SpectralElement


class MagnitudeSystem(BaseModel):
    """
    Class for a magnitude system.
    """

    name: MagnitudeSystemNames
    passbands: Annotated[
        list[str],
        Field(description=("List of passbands in the system.")),
    ]


class MagnitudeTransformPolynomial(BaseModel):
    """
    Class for magnitude transformation from one set of passbands to another that is best
    represented by a polynomial.
    """

    name: str
    from_passband: Annotated[
        str,
        Field(description="Passband to transform from."),
    ]
    to_passband: Annotated[
        str,
        Field(description="Passband to transform to."),
    ]
    polynomial_coefficients: Annotated[
        list[float],
        Field(
            description=(
                "Coefficients for the transformation. The "
                "number of coefficients depends on the transformation."
            )
        ),
    ]
    residual: Annotated[
        float,
        Field(description=("Residual of the transformation.")),
    ]

    @property
    def polynomial(self) -> np.polynomial.Polynomial:
        """
        Polynomial transformation.
        """
        return np.polynomial.Polynomial(self.polynomial_coefficients)


class MagnitudeTransformMatrix(BaseModel):
    """
    Class for magnitude transformation from one set of passbands to another that is best
    represented by a matrix.
    """

    name: str
    from_passbands: Annotated[
        list[str],
        Field(description="Passbands to transform from."),
    ]
    to_passbands: Annotated[
        list[str],
        Field(description="Passbands to transform to."),
    ]
    transformation_matrix: Annotated[
        list[list[float]],
        Field(
            description=(
                "Transformation matrix for the transformation. "
                "Shape should be (n_from, n_to)."
            )
        ),
    ]

    @property
    def array(self) -> np.ndarray:
        """
        Transformation matrix.
        """
        return np.array(self.transformation_matrix)


def _parse_transform_coefficients(
    input: Any,
) -> dict[tuple[str, str], MagnitudeTransformPolynomial]:
    """
    Parse the transform coefficients to ensure they are in the correct format.
    """
    if isinstance(input, str):
        return input

    if not isinstance(input, dict):
        raise ValueError(
            "Transform coefficients must be a dictionary of "
            "magnitude transformations."
        )
    return {
        tuple(k.split(",")) if isinstance(k, str) else k: v for k, v in input.items()
    }


class MagnitudeSystemTransform(BaseModel):
    """
    Class for magnitude system transformation from one single
    system to another.
    """

    name: str
    reference: Annotated[
        str,
        Field(
            description=(
                "Reference for the transformation, e.g., a " "paper or documentation."
            )
        ),
    ]
    from_system: MagnitudeSystem = Field(
        description="The magnitude system to transform from."
    )
    to_system: MagnitudeSystem = Field(
        description="The magnitude system to transform to."
    )
    transform_information: (
        Annotated[
            dict[tuple[str, str], MagnitudeTransformPolynomial],
            Field(
                description=(
                    "Coefficients for the transformation. The "
                    "number of coefficients depends on the transformation."
                )
            ),
            BeforeValidator(
                _parse_transform_coefficients,
            ),
        ]
        | Annotated[
            MagnitudeTransformMatrix,
            Field(
                description=(
                    "Transformation matrix for the transformation. "
                    "Shape should be (n_from, n_to)."
                )
            ),
        ]
    )


class PanStarrs1ToJohnsonCousinsMixin:
    """
    Implementation of the Pan-STARRS1 to Johnson-Cousins transformation.
    """

    def __call__(
        self, from_magnitudes: np.typing.ArrayLike, ps1_band_for_V: str = "gp1"
    ) -> np.ndarray:
        """
        Transform Pan-STARRS1 magnitudes to Johnson-Cousins magnitudes.

        Parameters
        ----------
        from_magnitudes : np.ArrayLike
            Pan-STARRS1 magnitudes to transform. The shape should either be (6,)
            or (n, 6), where n is the number of magnitudes in the Pan-STARRS1 system
            you wish to transform to Johnson-Cousins system.

        """
        # Convert to numpy array
        from_magnitudes = np.asarray(from_magnitudes)

        # The Pan-STARRS1 transformation is a polynomial of order 2
        # in which the independent variable, which they call "x",
        # is g_p1 - r_p1.
        from_band_index = {
            band: index for index, band in enumerate(self.from_system.passbands)
        }
        color = from_magnitudes[..., 0] - from_magnitudes[..., 1]
        if len(from_magnitudes.shape) == 1:
            from_magnitudes = from_magnitudes[np.newaxis, :]
        to_mags = np.zeros(
            from_magnitudes.shape[:-1] + (len(self.to_system.passbands),)
        )
        for index, jc_band in enumerate(self.to_system.passbands):
            to_band_name = jc_band
            match to_band_name:
                case "B":
                    from_band_name = "gp1"

                case "V":
                    # There are three options in the Pan-STARRS1 paper
                    # for the V band: g_p1, r_p1, and w_p1.
                    from_band_name = ps1_band_for_V
                case "Rc":
                    from_band_name = "rp1"
                case "Ic":
                    from_band_name = "ip1"
                case _:  # pragma: no cover
                    raise ValueError(
                        f"Unknown band name {to_band_name} for "
                        f"transformation from Pan-STARRS1 to Johnson-Cousins."
                    )

            to_mags[..., index] = (
                self.transform_information[(from_band_name, to_band_name)].polynomial(
                    color
                )
                + from_magnitudes[..., from_band_index[from_band_name]]
            )
        return np.squeeze(to_mags)


class MatrixTransformMixin:
    """
    Implementation of a matrix-based magnitude transformation.
    """

    def __call__(self, from_magnitudes: np.typing.ArrayLike) -> np.ndarray:
        """
        Transform magnitudes using a matrix transformation.

        Parameters
        ----------
        from_magnitudes : np.ArrayLike
            Magnitudes to transform. The shape should be (n, m), where n is the number
            of magnitudes and m is the number of passbands in the from_system. There
            must be an entry for each passband in the from_system, even if all of the
            entries for a particular passband are zero.
        """
        # Convert to numpy array
        from_magnitudes = np.asarray(from_magnitudes)

        # Ensure the input shape matches the expected number of passbands. The
        # +1 is for the constant term in the transforms
        if from_magnitudes.shape[0] != len(self.from_system.passbands) + 1:
            raise ValueError(
                f"Input shape {from_magnitudes.shape} does not match the number "
                f"of passbands in the from_system ({len(self.from_system.passbands)})."
            )

        # Perform the matrix multiplication
        return self.transform_information.array @ from_magnitudes


class PanStarrs1ToJohnsonCousins(
    PanStarrs1ToJohnsonCousinsMixin, MagnitudeSystemTransform
):
    """
    Class for transforming Pan-STARRS1 magnitudes to Johnson-Cousins magnitudes.
    """

    @classmethod
    def load(cls) -> "PanStarrs1ToJohnsonCousins":
        """
        Load the Pan-STARRS1 to Johnson-Cousins transformation from a file.
        """
        # Load the transformation from a file
        path = Path(get_pkg_data_path("data/PS1_to_JC.json"))
        return cls.model_validate_json(path.read_text())


class USNOPrimeToSDSSDR7(MatrixTransformMixin, MagnitudeSystemTransform):
    """
    Class for transforming USNO Prime magnitudes to SDSS DR7 magnitudes.
    """

    @classmethod
    def load(cls) -> "USNOPrimeToSDSSDR7":
        """
        Load the USNO Prime to SDSS DR7 transformation from a file.
        """
        # Load the transformation from a file
        path = Path(get_pkg_data_path("data/USNO_SDSS_to_SDSS_DR7.json"))
        return cls.model_validate_json(path.read_text())


def transform_apass_bands(table, apply_sdssdr7_transform: bool = False):
    """
    A function for transforming from the native APASS bassbands to Johnson/Cousins
    R and I bands.

    Parameters
    ----------
    table : `astropy.table.Table`
        Table of catalog information. This is assumed to be in "passband" format
    """
    # Putting this here to avoid a circular import
    from .magnitude_transforms import filter_transform

    # Set up for input to Jester transform
    use_columns = dict(
        g="mag_SG",
        r="mag_SR",
        i="mag_SI",
    )

    # The Jester transformation really for transformation from the ugriz on the
    # 2.5m SDSS telescope to the Johnson-Cousins system.
    #
    # APASS is using u'g'r'i'z' (well, only g'r'i' in APASS) and so I *think*
    # that the right thing to do is to transform the APASS g'r'i' to the SDSS DR7
    # system, and then apply jester.
    if apply_sdssdr7_transform:
        # Load the transformation
        transform = USNOPrimeToSDSSDR7.load()

        # Get the g', r', i' columns
        gp = table["mag_SG"]
        rp = table["mag_SR"]
        ip = table["mag_SI"]
        up = np.zeros_like(gp)  # Placeholder for u' band, not used in APASS
        zp = np.zeros_like(gp)  # Placeholder for z' band, not used in APASS
        ones = np.ones_like(gp)  # Placeholder for the constant term
        # Prepare the data for the transformation
        input_mags = np.asarray([up, gp.data, rp.data, ip.data, zp, ones])
        # Transform the APASS g'r'i' to SDSS DR7
        sdss_mags = transform(input_mags)
        new_columns = {
            "mag_SG_tmp": sdss_mags[1, :],
            "mag_SR_tmp": sdss_mags[2, :],
            "mag_SI_tmp": sdss_mags[3, :],
        }
        for k, v in new_columns.items():
            table[k] = v

        use_columns = dict(
            g="mag_SG_tmp",
            r="mag_SR_tmp",
            i="mag_SI_tmp",
        )

    table["mag_RC"] = filter_transform(
        table,
        output_filter="R",
        **use_columns,
        transform="jester",
    )
    table["mag_IC"] = filter_transform(
        table,
        output_filter="I",
        **use_columns,
        transform="jester",
    )

    # Yes, this is dumb. You fix it if you want it less dumb.
    table["mag_R"] = table["mag_RC"]
    table["mag_I"] = table["mag_IC"]
    # The dumbness has ended for now

    # Remove the temporary columns if they were created
    if apply_sdssdr7_transform:
        del table["mag_SG_tmp", "mag_SR_tmp", "mag_SI_tmp"]

    return table


def transform_refcat2_bands(table):
    """
    A function for transforming from the native RefCat2 bassbands to Johnson/Cousins
    BVRI bands.

    Parameters
    ----------
    table : `astropy.table.Table`
        Table of catalog information. This is assumed to be in "passband" format
    """
    transform = PanStarrs1ToJohnsonCousins.load()

    # Prepare the data for the transformation
    num_rows = len(table)

    transform_input = np.array(
        [
            table["mag_SG"],
            table["mag_SR"],
            table["mag_SI"],
            table["mag_SZ"],
            [0] * num_rows,  # Placeholder for the y_p1 band
            [0] * num_rows,  # Placeholder for the w_p1 band
        ]
    )

    transformed_data = transform(
        transform_input.T,
        ps1_band_for_V="gp1",
    )

    table["mag_B"] = transformed_data[:, 0]
    table["mag_V"] = transformed_data[:, 1]
    table["mag_RC"] = transformed_data[:, 2]
    table["mag_IC"] = transformed_data[:, 3]

    # Yes, this is dumb. You fix it if you want it less dumb.
    # We should only have one name for RC and IC, not two...
    table["mag_R"] = table["mag_RC"]
    table["mag_I"] = table["mag_IC"]
    # The dumbness has ended for now

    return table



# The goal here is to both test code and the data files
from pathlib import Path

import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.utils.data import get_pkg_data_path
from pydantic import ValidationError

from stellarphot import apass_dr9, refcat2
from stellarphot.utils.magnitude_system_transforms import (
    MagnitudeSystem,
    MagnitudeSystemNames,
    MagnitudeSystemTransform,
    MagnitudeTransformPolynomial,
    PanStarrs1ToJohnsonCousins,
    USNOPrimeToSDSSDR7,
    transform_apass_bands,
    transform_refcat2_bands,
)


def _check_json_roundtrip(model):
    """
    Check that the model can be serialized to JSON and then deserialized
    back to the same model.
    """
    json_str = model.model_dump_json()
    new_model = model.model_validate_json(json_str)
    assert model == new_model


class TestMagnitudeSystem:
    def test_can_create_and_jc_file_is_good(self):
        jc_bands = ["U", "B", "V", "Rc", "Ic"]
        sys = MagnitudeSystem(name=MagnitudeSystemNames.JC, passbands=jc_bands)
        assert sys.name == MagnitudeSystemNames.JC
        assert sys.passbands == jc_bands

        # Check that Johnson-Cousins that we ship is as expected
        from_data = Path(
            get_pkg_data_path("data/JohnsonCousins.json", package="stellarphot.utils")
        )
        assert from_data.exists()
        assert sys == MagnitudeSystem.model_validate_json(from_data.read_text())

        # Check serialization roundtrip
        _check_json_roundtrip(sys)

    def test_no_unknown_systems(self):
        with pytest.raises(ValidationError):
            MagnitudeSystem(name="Unknown", passbands=["U", "B", "V", "R", "I"])

    def test_panstarrs1_system_def(self):
        ps1_file = Path(
            get_pkg_data_path("data/pan-starrs1.json", package="stellarphot.utils")
        )
        ps1 = MagnitudeSystem.model_validate_json(ps1_file.read_text())
        assert ps1.name == MagnitudeSystemNames.PANSTARRS1
        assert ps1.passbands == ["gp1", "rp1", "ip1", "zp1", "yp1", "wp1"]


class TestMagnitudeTransform:
    def test_make_transform(self):
        fake_coeff = [0.1, 0.2, 0.3]
        # This is NOT a real transform....just making sure that
        # the transform can be made
        my_transform = MagnitudeTransformPolynomial(
            name="test",
            from_passband="B",
            to_passband="gp1",
            polynomial_coefficients=fake_coeff,
            residual=0.1,
        )

        assert my_transform.name == "test"
        assert np.polynomial.Polynomial(fake_coeff) == my_transform.polynomial
        val = 2
        assert np.isclose(
            my_transform.polynomial(val), (0.1 + 0.2 * val + 0.3 * val**2)
        )

        # check roundtrip serialization
        _check_json_roundtrip(my_transform)


class TestMagnitudeSystemTransform:
    def test_make_transform(self):
        my_jc_system = MagnitudeSystem(
            name=MagnitudeSystemNames.JC, passbands=["B", "V"]
        )
        my_ps1_system = MagnitudeSystem(
            name=MagnitudeSystemNames.PANSTARRS1, passbands=["gp1", "rp1"]
        )

        fake_coeff = [0.1, 0.2, 0.3]
        my_mag_trans1 = MagnitudeTransformPolynomial(
            name="test",
            from_passband="B",
            to_passband="gp1",
            polynomial_coefficients=fake_coeff,
            residual=0.1,
        )

        my_mag_trans2 = MagnitudeTransformPolynomial(
            name="test",
            from_passband="V",
            to_passband="rp1",
            polynomial_coefficients=fake_coeff,
            residual=0.1,
        )

        my_transform = MagnitudeSystemTransform(
            name="test",
            reference="some paper I will never write",
            from_system=my_jc_system,
            to_system=my_ps1_system,
            transform_information={
                ("B", "gp1"): my_mag_trans1,
                ("V", "rp1"): my_mag_trans2,
            },
        )

        # Make sure we can access the transform coefficients as intended.
        assert my_transform.transform_information[("B", "gp1")] == my_mag_trans1
        assert my_transform.transform_information[("V", "rp1")] == my_mag_trans2

        # Serialize the model and make sure that the transform keys appear
        # as expected.
        serialized = my_transform.model_dump_json()
        assert "B,gp1" in serialized
        assert "V,rp1" in serialized

        # Check that the transform can roundtrip to json
        _check_json_roundtrip(my_transform)


class TestPanStarrs1ToJohnsonCousins:
    @pytest.mark.parametrize("setup_method", ["class", "load"])
    def test_transform(self, setup_method):
        if setup_method == "class":
            # Here we just read in the transform on disk rather than doing it manually.
            ps1_file = Path(
                get_pkg_data_path("data/PS1_to_JC.json", package="stellarphot.utils")
            )
            ps1_to_jc = PanStarrs1ToJohnsonCousins.model_validate_json(
                ps1_file.read_text()
            )
        else:
            ps1_to_jc = PanStarrs1ToJohnsonCousins.load()

        # Spot check a couple of transform relationships
        # Expected polynomials from the paper
        gp1_B_poly = np.polynomial.Polynomial([0.212, 0.556, 0.034])
        assert ps1_to_jc.transform_information[("gp1", "B")].polynomial == gp1_B_poly

        # Check that the transform works
        fake_gp1_mags = np.array(
            [
                [20.0, 19.0, 18.0, 17.0, 16.0, 15.0],
                [20.0, 19.0, 18.0, 17.0, 16.0, 15.0],
                [20.0, 19.0, 18.0, 17.0, 16.0, 15.0],
            ]
        )
        # Transform just a single row
        fake_jc_mags = ps1_to_jc(fake_gp1_mags[1, :])
        assert fake_jc_mags.shape == (4,)
        assert np.isclose(
            fake_jc_mags[0],
            gp1_B_poly(fake_gp1_mags[0, 0] - fake_gp1_mags[0, 1]) + fake_gp1_mags[0, 0],
        )

        # Transform it all
        fake_jc_mags = ps1_to_jc(fake_gp1_mags)
        assert fake_jc_mags.shape == (3, 4)
        assert np.isclose(
            fake_jc_mags[0, 0],
            gp1_B_poly(fake_gp1_mags[0, 0] - fake_gp1_mags[0, 1]) + fake_gp1_mags[0, 0],
        )


class TestCatalogTransforms:
    @pytest.mark.remote_data
    def test_apass_adds_RI(self):
        # Get some APASS data
        apass = apass_dr9(SkyCoord(0, 0, unit="degree"), radius="1 arcmin")
        assert "mag_R" not in apass.columns
        assert "mag_I" not in apass.columns

        # Transform the data to add R and I
        apass_trans = apass.passband_columns(
            ["R", "I"], transformer=transform_apass_bands
        )
        assert "mag_R" in apass_trans.columns
        assert "mag_I" in apass_trans.columns

    @pytest.mark.remote_data
    def test_apass_can_apply_usno(self):
        # Get some APASS data
        apass = apass_dr9(SkyCoord(0, 0, unit="degree"), radius="1 arcmin")
        assert "mag_R" not in apass.columns
        assert "mag_I" not in apass.columns

        # Transform the data to add R and I
        apass_trans = apass.passband_columns(
            ["R", "I"],
            transformer=transform_apass_bands,
            transformer_kwargs=dict(apply_sdssdr7_transform=True),
        )
        assert "mag_R" in apass_trans.columns
        assert "mag_I" in apass_trans.columns

    @pytest.mark.remote_data
    def test_refcats_adds_BVRI(self):
        # Get some APASS data
        refcat = refcat2(SkyCoord(0, 0, unit="degree"), radius="1 arcmin")
        assert "mag_B" not in refcat.columns
        assert "mag_V" not in refcat.columns
        assert "mag_R" not in refcat.columns
        assert "mag_I" not in refcat.columns

        # Transform the data to add B, V, R, and I
        refcat_trans = refcat.passband_columns(
            ["B", "V", "R", "I"], transformer=transform_refcat2_bands
        )
        assert "mag_B" in refcat_trans.columns
        assert "mag_V" in refcat_trans.columns
        assert "mag_R" in refcat_trans.columns
        assert "mag_I" in refcat_trans.columns

        # Make sure an error is raised when a unknown band is requested
        with pytest.raises(
            ValueError, match="Transformer did not add columns for passbands"
        ):
            refcat.passband_columns(["X"], transformer=transform_refcat2_bands)


class TestUSNOPrimeToSDSSDR7:
    def test_transform(self):
        # Check that the zero point for ug is correct
        inp_mag = np.asarray([1.39, 0.0, 0.0, 0.0, 0.0, 1.0])
        # This should produce this output:
        # u: same as u' so 1.39
        # g: product of -b_g and -zp_gr, so -0.0318
        # r: product of -b_r and -zp_ri, so -0.00735
        # i: product of -b_i and -zp_ri, so -0.00861
        # z: product of -b_z and -zp_iz, so 0.0027

        # This effectively checks the whole matrix
        expected = np.asarray([1.39, -0.0318, -0.00735, -0.00861, 0.0027, 1])
        usno_to_sdss = USNOPrimeToSDSSDR7.load()
        out_mag = usno_to_sdss(inp_mag)
        assert np.allclose(out_mag, expected)



import os
from copy import deepcopy
from pathlib import Path

import ipywidgets as ipw
import pytest
from astropy.units import Quantity
from ipyautoui.custom.iterable import ItemBox, ItemControl
from pydantic import ValidationError
from pydantic.alias_generators import to_snake

from stellarphot.settings import (
    Camera,
    LoggingSettings,
    Observatory,
    PartialPhotometrySettings,
    PassbandMap,
    PhotometryApertures,
    PhotometryOptionalSettings,
    PhotometryWorkingDirSettings,
    SavedSettings,
    SourceLocationSettings,
    settings_files,
    ui_generator,
)
from stellarphot.settings.constants import (
    TEST_CAMERA_VALUES,
    TEST_OBSERVATORY_SETTINGS,
    TEST_PASSBAND_MAP,
    TEST_PHOTOMETRY_SETTINGS,
)
from stellarphot.settings.custom_widgets import (
    ChooseOrMakeNew,
    Confirm,
    ReviewSettings,
    SaveStatus,
    SettingWithTitle,
    Spinner,
    _add_saving_to_widget,
)

# Make sure we only use copies of the test settings
TEST_CAMERA_VALUES = deepcopy(TEST_CAMERA_VALUES)
TEST_OBSERVATORY_SETTINGS = deepcopy(TEST_OBSERVATORY_SETTINGS)
TEST_PASSBAND_MAP = deepcopy(TEST_PASSBAND_MAP)
TEST_PHOTOMETRY_SETTINGS = deepcopy(TEST_PHOTOMETRY_SETTINGS)


# See test_settings_file.TestSavedSettings for a detailed description of what the
# following fixture does. In brief, it patches the settings_files.PlatformDirs class
# so that the user_data_dir method returns the temporary directory.
@pytest.fixture(autouse=True)
def fake_settings_dir(mocker, tmp_path):
    mocker.patch.object(
        settings_files.PlatformDirs, "user_data_dir", tmp_path / "stellarphot"
    )


class TestChooseOrMakeNew:
    """
    Class for testing the ChooseOrMakeNew widget.
    """

    def make_test_camera(self):
        """
        Make a camera with the default testing values and save it. This came
        up often enough to warrant its own method.
        """
        saved = SavedSettings()
        camera = Camera(**TEST_CAMERA_VALUES)
        saved.add_item(camera)
        # Return the camera that was saved in case the test wants to use it
        return camera

    def test_creation_without_type_raises_error(self):
        # Should raise an error if no type is provided
        with pytest.raises(TypeError):
            ChooseOrMakeNew()

    def test_creation_unknown_item_type_raises_error(self):
        # Should raise an error if an unknown item type is provided
        with pytest.raises(ValueError, match="Unknown item type unknown"):
            ChooseOrMakeNew("unknown")

    @pytest.mark.parametrize(
        "item_type",
        [
            "camera",
            "observatory",
            "passband_map",
            "Camera",
            "PassbandMap",
            "Observatory",
        ],
    )
    def test_creation(self, item_type):
        # Should create a widget with a dropdown and a button
        choose_or_make_new = ChooseOrMakeNew(item_type)
        assert choose_or_make_new._item_type_name == item_type

    def test_initial_configuration_with_no_items(self):
        # Should have a dropdown with one item, "Make new passband map"
        # using passband_map here to also test that underscores get converted to spaces
        choose_or_make_new = ChooseOrMakeNew("passband_map")
        assert len(choose_or_make_new._choose_existing.options) == 1
        assert choose_or_make_new._choose_existing.options[0] == (
            "Make new passband map",
            "none",
        )

    def test_make_new_makes_a_new_item(self):
        # We will test this with Camera, should work for the others too since the code
        # path is the same.

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Set the values for the new item
        choose_or_make_new._item_widget.value = TEST_CAMERA_VALUES

        # Simulate a click on the save button...
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()
        # No need for the confirmation dialog because we are not overwriting
        # anything.

        # making sure the widget state is properly updated
        assert choose_or_make_new._making_new is False

        # Check what we created using SavedSettings...
        saved = SavedSettings()
        cameras = saved.get_items("camera")
        assert len(cameras.as_dict) == 1
        assert list(cameras.as_dict.values())[0].model_dump() == TEST_CAMERA_VALUES

    @pytest.mark.parametrize(
        "item_type,setting",
        [
            ("camera", TEST_CAMERA_VALUES),
            ("observatory", TEST_OBSERVATORY_SETTINGS),
            ("passband_map", TEST_PASSBAND_MAP),
        ],
    )
    def test_make_new_with_existing_item_resets_value(self, item_type, setting):
        # When "make a new" item is selected the value of the widget should be
        # the same as when a widget of that type is created.

        # Make a camera widget just to get the value for a new item.
        choose_or_make_new = ChooseOrMakeNew(item_type)
        value_when_new = choose_or_make_new._item_widget.value.copy()

        # Make a camera
        saved = SavedSettings()
        item = choose_or_make_new._item_widget.model(**setting)
        saved.add_item(item)

        # Make a camera widget and select "Make new"
        choose_or_make_new = ChooseOrMakeNew(item_type)
        choose_or_make_new._choose_existing.value = "none"
        assert choose_or_make_new._item_widget.value == value_when_new

    def test_edit_requires_confirmation(self):
        # Should require confirmation if the item already exists
        self.make_test_camera()
        choose_or_make_new = ChooseOrMakeNew("camera")
        # the edit button should be displayed and the confirm widget should be hidden
        # note: display typically start as None or an empty string, so we just check
        # that it is not "none", which is what it will be set to when it is hidden.
        assert choose_or_make_new._edit_delete_container.layout.display != "none"
        assert choose_or_make_new._confirm_edit_delete.layout.display == "none"

        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()
        # The edit button should now be hidden
        assert choose_or_make_new._edit_delete_container.layout.display == "none"

        # The savebuttonbar should be displayed
        assert choose_or_make_new._item_widget.savebuttonbar.layout.display != "none"

        # Click on the save button
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()

        # The confirm dialog should be displayed
        assert choose_or_make_new._confirm_edit_delete.layout.display != "none"

        # The confirmation dialog should contain the word "replace"
        assert "replace" in choose_or_make_new._confirm_edit_delete.message.lower()

    def test_edit_item_saved_after_confirm(self):
        # Should save the item after confirmation
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()

        # Change a value in the camera so we can check that the new value is saved.
        new_gain = 2 * Quantity(TEST_CAMERA_VALUES["gain"])
        choose_or_make_new._item_widget.di_widgets["gain"].value = str(new_gain)
        # Simulate a click on the save button...
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()
        # Simulate a click on the confirm button...
        choose_or_make_new._confirm_edit_delete._yes.click()

        saved = SavedSettings()
        cameras = saved.get_items("camera")
        assert cameras.as_dict[TEST_CAMERA_VALUES["name"]].gain == new_gain

    def test_edit_item_not_saved_after_cancel(self):
        # Should not save the item after clicking the No button
        camera = self.make_test_camera()

        # Make an additional test camera so that the result is the *second* camera
        # in the list, not the first. The workflow being test is that there are
        # already two cameras, then the second one is selected (so that something is
        # selected) and then that second one is edited, but "no" is clicked on the
        # confirmation dialog.
        saved = SavedSettings()
        camera.name = "zzzz" + camera.name
        saved.add_item(camera)

        choose_or_make_new = ChooseOrMakeNew("camera")

        # There should be three choices in the dropdown -- the two cameras and
        # "Make new"
        assert len(choose_or_make_new._choose_existing.options) == 3

        # Choose the second camera
        choose_or_make_new._choose_existing.value = camera

        assert camera == choose_or_make_new.value

        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()

        # Change a value in the camera so we can check that the new value is saved.
        new_gain = 2 * Quantity(TEST_CAMERA_VALUES["gain"])
        choose_or_make_new._item_widget.di_widgets["gain"].value = str(new_gain)
        # Simulate a click on the save button...
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()
        # Simulate a click on the cancel button...
        choose_or_make_new._confirm_edit_delete._no.click()
        assert camera == choose_or_make_new.value

    def test_selecting_make_new_as_selection_works(self):
        # Should allow the user to select "Make new" as a selection

        # Make a camera so that we can test that selecting "Make new" works
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        # This is the value for the "Make new...." option
        choose_or_make_new._choose_existing.value = "none"

        # edit button should go away
        assert choose_or_make_new._edit_delete_container.layout.display == "none"

        # Widget should be enabled
        assert choose_or_make_new._item_widget.disabled is False

        # save button bar should be displayed
        assert choose_or_make_new._item_widget.savebuttonbar.layout.display != "none"

    def test_choosing_different_item_updates_display(self):
        # Should update the display when a different item is chosen
        saved = SavedSettings()
        observatory = Observatory(**TEST_OBSERVATORY_SETTINGS)
        saved.add_item(observatory)

        observatory2 = Observatory(**TEST_OBSERVATORY_SETTINGS)
        # Make sure this name sorts lower than the first observatory
        observatory2.name = "zzzz" + observatory2.name
        observatory2.elevation = "-200 m"

        saved.add_item(observatory2)

        choose_or_make_new = ChooseOrMakeNew("observatory")

        # There should be three choices in the dropdown
        assert len(choose_or_make_new._choose_existing.options) == 3

        # Make sure the correct first observatory is selected
        assert choose_or_make_new._choose_existing.value == observatory
        assert Observatory(**choose_or_make_new._item_widget.value) == observatory

        # Select the other observatory
        choose_or_make_new._choose_existing.value = observatory2

        # The item widget should now have the values of the second observatory
        assert Observatory(**choose_or_make_new._item_widget.value) == observatory2

    def test_passband_map_buttons_are_disabled_or_enabled(self):
        # When an existing PassbandMap is selected the add/remove buttons
        # for individual rows should not be displayed.
        saved = SavedSettings()
        passband_map = PassbandMap(**TEST_PASSBAND_MAP)
        saved.add_item(passband_map)

        choose_or_make_new = ChooseOrMakeNew("passband_map")

        # There is no great way to get to the ItemBox widget that contains and controls
        # the add/remove buttons, so we keep going down through widget children until we
        # get to an ItemBox and then check that the buttons are disabled.
        # Recursion is the easiest way to do that, so recurse we will..
        def find_item_box(top_widget):
            for kid in top_widget.children:
                if isinstance(kid, ItemBox):
                    return kid
                if hasattr(kid, "children"):
                    result = find_item_box(kid)
                    if result:
                        return result

        item_box = find_item_box(choose_or_make_new)
        assert item_box.add_remove_controls == ItemControl.none

        # Next, we will click the "Edit" button and check that the buttons are enabled.
        choose_or_make_new._edit_button.click()
        item_box = find_item_box(choose_or_make_new)
        assert item_box.add_remove_controls == ItemControl.add_remove

    def test_make_passband_map(self):
        # Make a passband map and save it, then check that it is in the dropdown
        saved = SavedSettings()
        passband_map = PassbandMap(**TEST_PASSBAND_MAP)
        saved.add_item(passband_map)

        # Should create a new passband map
        choose_or_make_new = ChooseOrMakeNew("passband_map")
        assert len(choose_or_make_new._choose_existing.options) == 2
        assert choose_or_make_new._choose_existing.options[0][0] == passband_map.name

    def test_no_edit_button_when_there_are_no_items(self):
        # Should not have an edit button when there are no items
        saved = SavedSettings()
        # Make sure there are no cameras
        assert len(saved.get_items("camera").as_dict) == 0

        choose_or_make_new = ChooseOrMakeNew("camera")
        assert choose_or_make_new._edit_delete_container.layout.display == "none"

    def test_edit_button_returns_after_making_new_item(self):
        # After making a new item the edit button should be displayed

        # There are no cameras so this puts us into the form to make a new one
        choose_or_make_new = ChooseOrMakeNew("camera")

        # Make sure the edit button is hidden
        assert choose_or_make_new._edit_delete_container.layout.display == "none"

        # Change a value in the camera so we can check that the new value is saved.
        choose_or_make_new._item_widget.value = TEST_CAMERA_VALUES
        # Simulate a click on the save button...
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()

        # The edit button should now be displayed
        assert choose_or_make_new._edit_delete_container.layout.display != "none"

        # Select Make a new camera again
        choose_or_make_new._choose_existing.value = "none"

        # The edit button should now be hidden
        assert choose_or_make_new._edit_delete_container.layout.display == "none"

        # Select the camera we made
        choose_or_make_new._choose_existing.value = Camera(**TEST_CAMERA_VALUES)

        # The edit button should now be displayed
        assert choose_or_make_new._edit_delete_container.layout.display != "none"

    def test_save_button_disabled_when_no_changes(self):
        # Immediately after the edit button has been clicked, the save button should be
        # disabled because no changes have been made yet.
        #
        # That should remain true even after the user makes a change, then saves it
        # and then edits again but has not yet made any changes.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()

        # The save button should be disabled
        assert choose_or_make_new._item_widget.savebuttonbar.bn_save.disabled

        # Edit a value in the camera so the save button is enabled
        new_gain = 2 * Quantity(TEST_CAMERA_VALUES["gain"])
        choose_or_make_new._item_widget.di_widgets["gain"].value = str(new_gain)

        # The save button should now be enabled
        assert not choose_or_make_new._item_widget.savebuttonbar.bn_save.disabled

        # Save the change
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()

        # Confirm the save
        choose_or_make_new._confirm_edit_delete._yes.click()

        # Edit again...
        choose_or_make_new._edit_button.click()

        # The save button should be disabled
        assert choose_or_make_new._item_widget.savebuttonbar.bn_save.disabled

    def test_revert_button_is_enabled_after_clicking_edit(self):
        # The revert button should be enabled after clicking the edit button so
        # the user can cancel the edit.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()

        # The revert button should be enabled
        assert not choose_or_make_new._item_widget.savebuttonbar.bn_revert.disabled

    def test_clicking_revert_button_cancels_edit(self):
        # Clicking the revert button should cancel the edit
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()

        # Simulate a click on the revert button...
        choose_or_make_new._item_widget.savebuttonbar.bn_revert.click()

        # The camera should not have been changed
        assert choose_or_make_new._item_widget.value == TEST_CAMERA_VALUES

        # The edit button should be displayed
        assert choose_or_make_new._edit_delete_container.layout.display != "none"

        # We should not be in editing mode anymore
        assert not choose_or_make_new._editing

    def test_revert_button_remains_enabled_with_invalid_value_and_actually_reverts(
        self,
    ):
        # The revert button should remain enabled if the value is invalid and reverting
        # should actually revert the value.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Simulate a click on the edit button...
        choose_or_make_new._edit_button.click()

        # Set an invalid value
        choose_or_make_new._item_widget.di_widgets["name"].value = ""

        # unsaved_changes should be true
        assert choose_or_make_new._item_widget.savebuttonbar.unsaved_changes

        # Make sure the change is really there
        assert choose_or_make_new._item_widget.value != TEST_CAMERA_VALUES

        # The revert button should be enabled
        assert not choose_or_make_new._item_widget.savebuttonbar.bn_revert.disabled

        # Save should still be disabled
        assert choose_or_make_new._item_widget.savebuttonbar.bn_save.disabled

        # Click the revert button
        choose_or_make_new._item_widget.savebuttonbar.bn_revert.click()

        # The camera should not have been changed
        assert choose_or_make_new._item_widget.value == TEST_CAMERA_VALUES

    def test_delete_button_click_displays_confirm_dialog(self):
        # Clicking "Delete" should display a confirmation dialog

        # Make an item to ensure an existing item is displayed when the widget is made
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        choose_or_make_new._delete_button.click()

        # Confirm dialog should be displayed
        assert choose_or_make_new._confirm_edit_delete.layout.display != "none"
        # edit/delete buttons should be hidden
        assert choose_or_make_new._edit_delete_container.layout.display == "none"
        # The confirm dialog should indicate you are deleting
        assert "delete" in choose_or_make_new._confirm_edit_delete.message.lower()

    @pytest.mark.parametrize("click_yes", [True, False])
    def test_delete_actions_after_confirmation(self, click_yes):
        # Clicking "Yes" in the confirmation dialog should delete the item
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        choose_or_make_new._delete_button.click()

        if click_yes:
            # Simulate a click on the "Yes" button in the confirm dialog
            choose_or_make_new._confirm_edit_delete._yes.click()

            # The camera should be gone
            saved = SavedSettings()
            cameras = saved.get_items("camera")
            assert len(cameras.as_dict) == 0
        else:
            # Simulate a click on the "No" button in the confirm dialog
            choose_or_make_new._confirm_edit_delete._no.click()

            # The camera should still be there
            saved = SavedSettings()
            cameras = saved.get_items("camera")
            assert len(cameras.as_dict) == 1
            assert list(cameras.as_dict.values())[0].model_dump() == TEST_CAMERA_VALUES

        # Regardless of the choice, the confirm dialog should be hidden
        assert choose_or_make_new._confirm_edit_delete.layout.display == "none"

        # We should also no longer be deleting
        assert not choose_or_make_new._deleting

        # If the user clicked "No" the edit/delete buttons should be displayed because
        # there is still an item to edit or delete.
        if not click_yes:
            assert choose_or_make_new._edit_delete_container.layout.display != "none"
        # If the user clicked "Yes" the edit/delete buttons should be hidden because
        # there is no item to edit or delete.
        else:
            assert choose_or_make_new._edit_delete_container.layout.display == "none"

    def test_correct_item_selected_after_delete_and_yes(self):
        # The correct item should be selected after an item is deleted and the user
        # clicks "Yes" in the confirmation dialog.
        # Make two cameras
        self.make_test_camera()
        saved = SavedSettings()
        camera2 = Camera(**TEST_CAMERA_VALUES)
        camera2.name = "zzzz" + camera2.name
        saved.add_item(camera2)

        choose_or_make_new = ChooseOrMakeNew("camera")
        # Select the first camera...
        choose_or_make_new._choose_existing.value = saved.cameras.as_dict[
            TEST_CAMERA_VALUES["name"]
        ]
        # ...and delete it
        choose_or_make_new._delete_button.click()
        choose_or_make_new._confirm_edit_delete._yes.click()
        # Are the edit/delete buttons being shown?
        assert choose_or_make_new._edit_delete_container.layout.display != "none"
        # Is the correct camera selected?
        assert choose_or_make_new._choose_existing.value == camera2

    @pytest.mark.parametrize("click_yes", [True, False])
    def test_making_new_with_same_name_as_existing(self, click_yes):
        # Making a new item with the same name as an existing item should require
        # confirmation
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")

        # Click the "Make new" button
        choose_or_make_new._choose_existing.value = "none"

        # set the item widget value to the existing camera with double the gain
        camera_values = deepcopy(TEST_CAMERA_VALUES)
        camera_values["gain"] = 2 * Quantity(camera_values["gain"])
        choose_or_make_new._item_widget.value = Camera(**camera_values)

        # Simulate a click on the save button...
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()

        # The confirm dialog should be displayed
        assert choose_or_make_new._confirm_edit_delete.layout.display != "none"
        # The confirm dialog should contain the word "replace"
        assert "replace" in choose_or_make_new._confirm_edit_delete.message.lower()

        if click_yes:
            # ...click confirm
            choose_or_make_new._confirm_edit_delete._yes.click()
            expected_camera = Camera(**camera_values)
        else:
            # ...click cancel
            choose_or_make_new._confirm_edit_delete._no.click()
            expected_camera = Camera(**TEST_CAMERA_VALUES)

        # Check the value of the camera
        chosen_cam = Camera(**choose_or_make_new._item_widget.value)
        assert chosen_cam == choose_or_make_new._choose_existing.value

        assert chosen_cam == expected_camera

    def test_weird_sequence_of_no_clicks(self):
        # This is a regression test for #320
        # Make a camera
        self.make_test_camera()

        # Make a new camera...
        choose_or_make_new = ChooseOrMakeNew("camera")
        choose_or_make_new._choose_existing.value = "none"
        choose_or_make_new._item_widget.value = TEST_CAMERA_VALUES

        # ...click save...
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()
        # ...but say no to confirm
        choose_or_make_new._confirm_edit_delete._no.click()

        # Check that we are back to our selected camera
        assert choose_or_make_new._choose_existing.value == Camera(**TEST_CAMERA_VALUES)

        # Click delete...
        choose_or_make_new._delete_button.click()
        # ...but say no to confirm...
        choose_or_make_new._confirm_edit_delete._no.click()

        # ...the edit/delete buttons should be displayed
        assert choose_or_make_new._edit_delete_container.layout.display != "none"

    @pytest.mark.parametrize("hideable", [True, False])
    def test_details_hideable_or_not(self, hideable):
        # The details should be hideable if requested

        # Make a camera so that the details can be hidden
        self.make_test_camera()
        choose_or_make_new = ChooseOrMakeNew("camera", details_hideable=hideable)

        # New UI element should be in the big box
        assert (
            choose_or_make_new._show_details_ui
            in choose_or_make_new._choose_detail_container.children
        )

        if hideable:
            # The "show/hide details" widget should be visible
            assert choose_or_make_new._show_details_ui.layout.display != "none"
        else:
            # The "show/hide details" widget should not be visible
            assert choose_or_make_new._show_details_ui.layout.display == "none"

        # Set "show details" box to unchecked
        choose_or_make_new._show_details_ui.value = False

        # If hideable then details should not be displayed
        if hideable:
            assert choose_or_make_new._details_box.layout.display == "none"
        else:
            assert choose_or_make_new._details_box.layout.display != "none"

        # Settings show_details back to true should display details again,
        # regardless of hideable
        choose_or_make_new._show_details_ui.value = True
        assert choose_or_make_new._details_box.layout.display != "none"

    def test_details_hideable_plays_nicely_with_new_item(self):
        # The details should be hideable and should play nicely with making a new item
        # Make an item so that selecting "Make new" will count as a value
        # change later.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera", details_hideable=True)

        assert choose_or_make_new._details_box.layout.display != "none"

        # Set "show details" box to unchecked so details are hidden
        choose_or_make_new._show_details_ui.value = False

        # Check that details are hidden
        assert choose_or_make_new._details_box.layout.display == "none"

        # Select make a new camera
        choose_or_make_new._choose_existing.value = "none"

        # The details should be shown now
        assert choose_or_make_new._details_box.layout.display != "none"

        # The "show/hide details" widget should be hidden
        assert choose_or_make_new._show_details_ui.layout.display == "none"

        # Making a new camera and saving it should bring back, and respect,
        # the hideable details
        new_camera = deepcopy(TEST_CAMERA_VALUES)
        new_camera["name"] = "new camera"
        choose_or_make_new._item_widget.value = new_camera
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()

        # "show details" should be visible
        assert choose_or_make_new._show_details_ui.layout.display != "none"

        # Details should be hidden
        assert choose_or_make_new._details_box.layout.display == "none"
        assert not choose_or_make_new._show_details_ui.value

    def test_details_hideable_not_set_preserves_old_behavior(self):
        # The details should not be hidden if hideable is false
        # Make a camera
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera", details_hideable=False)

        # New UI element should not be in the big box
        assert choose_or_make_new._show_details_ui.layout.display == "none"

        # Details should be displayed
        assert choose_or_make_new._details_box.layout.display != "none"

        # Set "show details" box to unchecked
        choose_or_make_new._show_details_ui.value = False

        # Details should still be displayed
        assert choose_or_make_new._details_box.layout.display != "none"

    @pytest.mark.parametrize("show_detail_state", [True, False])
    def test_details_hideable_cancel_making_new_restores_state(self, show_detail_state):
        # The details should be hideable and should play nicely with making a new item
        # Make an item so that selecting "Make new" will count as a value
        # change later.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera", details_hideable=True)

        assert choose_or_make_new._details_box.layout.display != "none"

        # Note whether the "show details" box is checked or not
        show_state = show_detail_state
        choose_or_make_new._show_details_ui.value = show_state

        # Select make a new camera
        choose_or_make_new._choose_existing.value = "none"

        # The details should be shown now
        assert choose_or_make_new._details_box.layout.display != "none"

        # The "show/hide details" widget should be hidden
        assert choose_or_make_new._show_details_ui.layout.display == "none"

        # Making a new camera and then cancelling should restore the state
        choose_or_make_new._item_widget.value = TEST_CAMERA_VALUES
        choose_or_make_new._item_widget.savebuttonbar.bn_save.click()
        choose_or_make_new._confirm_edit_delete._no.click()

        # "show details" should be visible
        assert choose_or_make_new._show_details_ui.layout.display != "none"

        # Details should match prior state
        assert choose_or_make_new._show_details_ui.value == show_state

    def test_chooser_has_value(self):
        # The chooser should have a value
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera")
        assert choose_or_make_new.value == Camera(**TEST_CAMERA_VALUES)

    def test_details_can_be_hidden(self):
        # Make sure that item details can be hidden programmatically.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera", details_hideable=True)

        # Check that the details start out displayed
        assert choose_or_make_new._details_box.layout.display != "none"
        assert choose_or_make_new.display_details

        # Hide the details
        choose_or_make_new.display_details = False

        # Check that the details are hidden
        assert choose_or_make_new._details_box.layout.display == "none"

    def test_details_visibilty_cannot_be_changed_when_not_hideable(self):
        # Make sure that item details cannot be hidden programmatically when hideable
        # is False.
        self.make_test_camera()

        choose_or_make_new = ChooseOrMakeNew("camera", details_hideable=False)

        # Check that the details start out displayed
        assert choose_or_make_new._details_box.layout.display != "none"
        # Value should be None in this case
        assert choose_or_make_new.display_details is None

        # Hide the details
        choose_or_make_new.display_details = False
        # Value should still be None in this case
        assert choose_or_make_new.display_details is None

        # Check that the details are still displayed
        assert choose_or_make_new._details_box.layout.display != "none"


class TestConfirm:
    def test_initial_value(self):
        # Should initialize with value of None
        confirm = Confirm()
        assert confirm.value is None

    def test_initial_display(self):
        # Should initialize not displayed
        confirm = Confirm()
        assert confirm.layout.display == "none"

    def test_message(self):
        # Should initialize with a message
        confirm = Confirm(message="Hello")
        assert confirm.message == "Hello"
        # message should be settable
        confirm.message = "Goodbye"
        assert confirm.message == "Goodbye"

    def test_widget_structure(self):
        # Should have a message, yes, and no buttons
        confirm = Confirm()
        assert isinstance(confirm.children[0], ipw.HTML)
        assert isinstance(confirm.children[1], ipw.Button)
        assert isinstance(confirm.children[2], ipw.Button)

    @pytest.mark.parametrize("other_widget", [None, ipw.Dropdown()])
    def test_show(self, other_widget):
        # Should display when show is called
        confirm = Confirm(widget_to_hide=other_widget)
        if other_widget:
            assert other_widget.layout.display != "none"
        confirm.show()
        assert confirm.layout.display == "flex"
        if other_widget:
            assert other_widget.layout.display == "none"

    @pytest.mark.parametrize("yes_or_no", ["yes", "no"])
    @pytest.mark.parametrize("other_widget", [None, ipw.Dropdown()])
    def test_clicking_yes_or_no(self, yes_or_no, other_widget):
        # Should set value to True when yes is "clicked"
        confirm = Confirm(widget_to_hide=other_widget)
        # These are the handlers for a click on yes or no -- it takes an argument but we
        #  don't use it. The widget API requires that the function be able to take an
        # argument.
        if yes_or_no == "yes":
            confirm._yes.click()
        else:
            confirm._no.click()

        assert confirm.value == (yes_or_no == "yes")
        assert confirm.layout.display == "none"
        if other_widget:
            assert other_widget.layout.display != "none"


class TestSettingWithTitle:
    def test_title_format(self):
        # Check that the title is formatted with the requested heading
        # level and that there is, initially, no decoration.
        # Make a Camera with default testing settings
        camera = ui_generator(Camera)

        # Choosing something other than the default
        header_level = 5

        # Make a SettingWithTitle with the camera
        plain_title = "I am a camera"
        camera_title = SettingWithTitle(plain_title, camera, header_level=header_level)

        assert (
            camera_title.title.value
            == f"<h{header_level}>{plain_title}</h{header_level}>"
        )

    def test_title_decoration_plain_autoui(self):
        # Check that the title has the correct decoration given the
        # state of the settings widget.

        # Make a camera ui
        camera = ui_generator(Camera)
        plain_title = "I am a camera"

        # Make a SettingWithTitle with the camera
        camera_title = SettingWithTitle(plain_title, camera)

        # At the moment the title should not be decorated
        assert SaveStatus.SETTING_IS_SAVED not in camera_title.title.value
        assert SaveStatus.SETTING_NOT_SAVED not in camera_title.title.value

        # There should also be no unsaved changes at the moment
        assert not camera_title._autoui_widget.savebuttonbar.unsaved_changes

        # Manually set the value of the gain, which should make the trait
        # unsaved_changes True...
        camera_title._autoui_widget.di_widgets["gain"].value = str(
            2 * TEST_CAMERA_VALUES["gain"]
        )

        # ...and if unsaved_changes is True the "not saved" indicator should be present
        assert SaveStatus.SETTING_NOT_SAVED in camera_title.title.value

        # Click save...
        camera_title._autoui_widget.savebuttonbar.bn_save.click()

        # ... and check that the title is decorated with the "saved" indicator
        assert SaveStatus.SETTING_IS_SAVED in camera_title.title.value

        # Finally, click the save button and the title should be decorated with
        # the saved indication.
        camera_title._autoui_widget.savebuttonbar.bn_save.click()
        assert SaveStatus.SETTING_IS_SAVED in camera_title.title.value

    @pytest.mark.parametrize("accept_edits", [True, False])
    def test_title_decoration_choose_or_make_new_editing(self, accept_edits):
        # Check that the title has the correct decoration given the
        # state of the settings widget.

        # Make an item so there is something to select
        saved_setting = SavedSettings()
        saved_setting.add_item(Camera(**TEST_CAMERA_VALUES))

        camera = ChooseOrMakeNew(Camera.__name__)
        plain_title = camera._title.value

        # Make a SettingWithTitle with the camera
        camera_title = SettingWithTitle(plain_title, camera)

        # At the moment the title should not be decorated
        assert SaveStatus.SETTING_IS_SAVED not in camera_title.title.value
        assert SaveStatus.SETTING_NOT_SAVED not in camera_title.title.value

        # There should also be no unsaved changes at the moment
        assert not camera_title._autoui_widget.savebuttonbar.unsaved_changes

        # Click the edit button
        camera_title._widget._edit_button.click()

        # Edit a value in the camera so the save button is enabled
        new_gain = 2 * Quantity(TEST_CAMERA_VALUES["gain"])
        camera_title._widget._item_widget.di_widgets["gain"].value = str(new_gain)

        assert SaveStatus.SETTING_NOT_SAVED in camera_title.title.value

        # Click the save button....
        camera_title._widget._item_widget.savebuttonbar.bn_save.click()

        # ...should still be unsaved pending confirmation...
        assert SaveStatus.SETTING_NOT_SAVED in camera_title.title.value

        # ...now confirm or deny the save...
        if accept_edits:
            camera_title._widget._confirm_edit_delete._yes.click()
        else:
            camera_title._widget._confirm_edit_delete._no.click()

        # ...and the title should be decorated with the saved indication.
        assert SaveStatus.SETTING_IS_SAVED in camera_title.title.value


SETTING_CLASSES = [
    Camera,
    Observatory,
    PassbandMap,
    PhotometryApertures,
    PhotometryOptionalSettings,
    SourceLocationSettings,
    LoggingSettings,
]


def _to_space(name: str) -> str:
    return name.replace("_", " ")


class TestReviewSettings:
    """
    Test of the magical ReviewSettings widget.
    """

    # This auto-used fixture changes the working directory to the temporary directory
    # and then changes back to the original directory after the test is done.
    @pytest.fixture(autouse=True)
    def change_to_tmp_dir(self, tmp_path):
        original_dir = os.getcwd()
        os.chdir(tmp_path)
        # Yielding here is important. It means that when the test is done, the remainder
        # of the function will be executed. This is important because the test is run in
        # a temporary directory and we want to change back to the original directory
        # when the test is done.
        yield
        os.chdir(original_dir)

    @pytest.mark.parametrize("container_type", ["tabs", "accordion"])
    def test_creation_no_saved_settings(self, container_type):
        # Check creation and names of tab when there are no saved settings
        # and just one type of setting.
        for setting_class in SETTING_CLASSES:
            wd_settings = PhotometryWorkingDirSettings()

            # Remove any existing settings files that may have been saved in earlier
            # iterations of the loop.
            p = Path(".")
            (p / wd_settings.partial_settings_file).unlink(missing_ok=True)
            (p / wd_settings.settings_file).unlink(missing_ok=True)
            review_settings = ReviewSettings([setting_class], style=container_type)

            assert (
                _to_space(to_snake(setting_class.__name__))
                in review_settings._container.titles[0]
            )
            if container_type == "tabs":
                assert isinstance(review_settings._container, ipw.Tab)
            else:
                assert isinstance(review_settings._container, ipw.Accordion)

            # What happens next depends on whether the setting can be created from
            # default values or not.
            try:
                setting_class()
            except ValidationError:
                created_from_defaults = False
            else:
                created_from_defaults = True

            if created_from_defaults:
                # Test whether the setting, which was able to be created from default
                # values of the fields, is in the partial_settings.
                wd_settings.load()
                snake_name = to_snake(setting_class.__name__)
                assert (
                    getattr(wd_settings.partial_settings, snake_name) == setting_class()
                )
            else:
                with pytest.raises(
                    ValueError,
                    match="Settings file photometry_settings.json does not exist",
                ):
                    wd_settings.load()

    def test_creation_with_saved_settings(self):
        # Check creation and names of tab when there are saved settings
        # and just one type of setting.

        saveable_types = ChooseOrMakeNew._known_types
        for setting_class in SETTING_CLASSES:
            if setting_class.__name__ not in saveable_types:
                # The rest of this only makes sense if the setting class is saveable
                continue

            # Make a setting
            saved = SavedSettings()
            # Make an instance of the class
            snake_name = to_snake(setting_class.__name__)
            item = setting_class.model_validate(TEST_PHOTOMETRY_SETTINGS[snake_name])
            # Save the instance to saved settings file
            saved.add_item(item)

            # Make the review widget. It should auto-populate with the item with just
            # created because that is the only item of that type.
            review_settings = ReviewSettings([setting_class])
            model_dict = review_settings._container.children[0]._autoui_widget.value
            assert setting_class(**model_dict) == item

            # This setting was made automatically by the widget, so the title of the
            # container should end with SaveStatus.SETTING_SHOULD_BE_REVIEWED
            assert (
                SaveStatus.SETTING_SHOULD_BE_REVIEWED
                in review_settings._container.titles[0]
            )

            # The item setting should also have been saved to the working directory
            wd_settings = PhotometryWorkingDirSettings()
            loaded_settings = wd_settings.load()
            assert getattr(loaded_settings, snake_name) == item

    def test_creation_with_saved_working_dir_settings(self):
        # Check that when there is a saved item in the working directory
        # the badge attached to the name is "needs review"
        photometry_apertures = TEST_PHOTOMETRY_SETTINGS["photometry_apertures"]
        wk_dir = PhotometryWorkingDirSettings()
        partial_settings = PartialPhotometrySettings(
            photometry_apertures=photometry_apertures
        )
        wk_dir.save(partial_settings)

        # Create the review widget
        review_settings = ReviewSettings(
            [PhotometryOptionalSettings, PhotometryApertures]
        )

        assert (
            SaveStatus.SETTING_SHOULD_BE_REVIEWED
            in review_settings._container.titles[1]
        )

    def test_selecting_tab_updates_title(self):
        # Check that selecting a tab updates the title of the widget to end with the
        # appropriate indicator.
        # Save a camera and an observatory
        saved = SavedSettings()
        camera = Camera(**TEST_CAMERA_VALUES)
        saved.add_item(camera)
        observatory = Observatory(**TEST_OBSERVATORY_SETTINGS)
        saved.add_item(observatory)

        # Create the review widget
        review_settings = ReviewSettings([Camera, Observatory])

        # Select the Observatory tab
        review_settings._container.selected_index = 1

        assert SaveStatus.SETTING_IS_SAVED in review_settings._container.titles[1]

    def test_conflict_between_saved_and_working_dir(self):
        # Check that the expected error is raised if the value of a saved setting
        # like a camera conflicts with the value of the same setting in the working
        # directory.

        # Make a camera for the saved settings (i.e. those in user's settings directory)
        saved = SavedSettings()
        camera = Camera(**TEST_CAMERA_VALUES)
        saved.add_item(camera)

        # Make a camera for the working directory with different gain than the first
        # camera
        wd_settings = PhotometryWorkingDirSettings()
        wd_camera = Camera(**TEST_CAMERA_VALUES)
        wd_camera.gain = 2 * wd_camera.gain
        wd_settings.save(PartialPhotometrySettings(camera=wd_camera))

        # Create the review widget and check that the error is raised
        with pytest.raises(
            ValueError, match="The camera setting saved in the working directory is not"
        ):
            ReviewSettings([Camera])

    def test_error_when_setting_has_no_saved_or_default_setting(self):
        # Make a review widget with a setting that has no saved or default setting
        # like Camera and check that an error is raised.

        review_settings = ReviewSettings([Camera])

        assert SaveStatus.SETTING_NOT_SAVED in review_settings._container.titles[0]

    def test_setting_selected_item_to_none_does_not_save(self):
        # The intent of this is to test an edge case where the user selects an item
        # from the dropdown and then selects "None" from the dropdown. The selected
        # item should be saved, but there should not be another save when None is
        # selected.
        saved = SavedSettings()
        camera = Camera(**TEST_CAMERA_VALUES)
        saved.add_item(camera)
        wd_settings = PhotometryWorkingDirSettings()
        wd_set_path = Path(wd_settings.partial_settings_file)
        review_settings = ReviewSettings([Camera])

        # Get the modification time of the working directory settings file
        mtime = wd_set_path.stat().st_mtime

        assert (
            review_settings._setting_widgets[0]._widget._choose_existing.value == camera
        )
        review_settings._setting_widgets[0]._widget._choose_existing.value = "none"

        # Check that the working directory settings file has not been modified
        assert wd_set_path.stat().st_mtime == mtime

    def test_clicking_tab_with_already_saved_settings_updates_badge(self):
        # Check that if there are already settings saved to the working directory
        # and the user clicks on the tab, the badge is updated to reflect that the
        # setting has been saved.

        # To set this up we need to save settings to the working directory AND to the
        # saved user settings.
        wd_settings = PhotometryWorkingDirSettings()
        camera = Camera(**TEST_CAMERA_VALUES)
        observatory = Observatory(**TEST_OBSERVATORY_SETTINGS)
        passbands = PassbandMap(**TEST_PASSBAND_MAP)
        wd_settings.save(
            PartialPhotometrySettings(
                camera=camera, observatory=observatory, passband_map=passbands
            )
        )
        saved = SavedSettings()
        saved.add_item(camera)
        saved.add_item(observatory)
        saved.add_item(passbands)

        # Make the review widget
        review_settings = ReviewSettings([Camera, Observatory, PassbandMap])
        for title in review_settings._container.titles:
            assert SaveStatus.SETTING_SHOULD_BE_REVIEWED in title

        # Click on each tab, starting with the last one, to make sure a change in
        # the selected value happens to trigger the observers.
        for i in range(2, -1, -1):
            review_settings._container.selected_index = i

        for title in review_settings._container.titles:
            assert SaveStatus.SETTING_IS_SAVED in title

    def test_clicking_tab_marks_green_when_all_saved(self):
        # Check that if there are already settings saved to the working directory
        # and the user clicks on the tab, the badge is updated to reflect that the
        # setting has been saved.

        # To set this up we need to save settings to the working directory AND to the
        # saved user settings.
        wd_settings = PhotometryWorkingDirSettings()
        camera = Camera(**TEST_CAMERA_VALUES)
        observatory = Observatory(**TEST_OBSERVATORY_SETTINGS)
        passbands = PassbandMap(**TEST_PASSBAND_MAP)
        wd_settings.save(
            PartialPhotometrySettings(
                camera=camera, observatory=observatory, passband_map=passbands
            )
        )
        saved = SavedSettings()
        saved.add_item(camera)
        saved.add_item(observatory)
        saved.add_item(passbands)

        review_settings = ReviewSettings(
            [
                Camera,
                Observatory,
                PassbandMap,
                PhotometryApertures,
                SourceLocationSettings,
                PhotometryOptionalSettings,
                LoggingSettings,
            ]
        )

        num_tabs = len(review_settings._container.children)

        for title in review_settings._container.titles:
            assert SaveStatus.SETTING_SHOULD_BE_REVIEWED in title

        # Click on each tab, starting with the last one, to make sure a change in
        # the selected value happens to trigger the observers.
        for i in range(num_tabs - 1, -1, -1):
            review_settings._container.selected_index = i

        for title in review_settings._container.titles:
            assert SaveStatus.SETTING_IS_SAVED in title

    def test_loading_saved_source_locations_to_ui(self):
        # Check that saved source locations are loaded into the UI
        wd_set = PhotometryWorkingDirSettings()
        source_locations = SourceLocationSettings()
        wd_set.save(
            PartialPhotometrySettings(source_location_settings=source_locations)
        )

        review_settings = ReviewSettings([SourceLocationSettings])

        # Check that the saved source locations are in the UI
        assert (
            review_settings._setting_widgets[0]._autoui_widget.value
            == source_locations.model_dump()
        )
        assert (
            review_settings._setting_widgets[0]
            ._autoui_widget.di_widgets["source_list_file"]
            .selected_filename
            == source_locations.source_list_file
        )

    def test_getting_settings_with_nothing_saved(self):
        # Check that when we create the object with no saved settings we get an empty
        # partial settings object.
        review_settings = ReviewSettings([Camera])
        assert review_settings.current_settings == PartialPhotometrySettings()

    def test_selecting_table_without_saved_setting_sets_proper_badge(self):
        # Test that selecting a tab with a saved setting sets a proper "NOT SAVED" badge
        review_settings = ReviewSettings([PhotometryApertures, Camera])
        review_settings._container.selected_index = 1
        assert SaveStatus.SETTING_NOT_SAVED in review_settings._container.titles[1]


def test_add_saving_with_unrecognized_widget():
    # Check that an error is raised if a widget is added to the saving list
    # that is not recognized.
    with pytest.raises(ValueError, match="is not a recognized type of widget"):
        _add_saving_to_widget(ipw.Dropdown())


class TestSpinner:
    def test_create_spinner(self):
        # Test we can create a spinner and set a message
        spinner = Spinner()
        assert spinner.layout.display == "none"

        message = "Hello world"
        spinner = Spinner(message=message)
        assert spinner.message == message

        # Just make sure there is some spinner file
        assert spinner.spinner_file != ""

    def test_start_stop_spinner(self):
        # Test we can start and stop the spinner
        spinner = Spinner()
        spinner.start()
        assert spinner.layout.display == "flex"

        spinner.stop()
        assert spinner.layout.display == "none"

    def test_settings_spinner_file(self, tmp_path):
        # Test that we can set the spinner file. The file we
        # create here is not valid svg. We just make sure that
        # the file is read.

        spinner_file = tmp_path / "fake_spinner.svg"
        spinner_contents = "This is not a valid svg file"
        spinner_file.write_text(spinner_contents)

        spinner = Spinner(spinner_file=spinner_file)
        assert spinner.spinner_file == spinner_file
        assert spinner.children[1].value == spinner_contents



import warnings

import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import CCDData
from astropy.stats import gaussian_sigma_to_fwhm
from astropy.table import QTable
from astropy.utils.data import get_pkg_data_path
from astropy.utils.exceptions import AstropyUserWarning

from stellarphot import SourceListData
from stellarphot.photometry import compute_fwhm, fast_fwhm_from_image, source_detection
from stellarphot.photometry.tests.fake_image import FakeCCDImage, FakeImage
from stellarphot.settings.models import FwhmMethods

# Make sure the tests are deterministic by using a random seed
SEED = 5432985


@pytest.mark.parametrize("units", [u.pixel, None])
def test_compute_fwhm(units):
    fake_image = FakeImage(seed=SEED)
    sources = fake_image.sources
    if units is not None:
        # It turns out having a unit on a column is not the same as
        # things in the column having units. The construct below ensures
        # that the source table values have units.
        # Do not try: sources['x_mean'] = sources['x_mean'] * units
        # Turns out individual values do NOT have units in that case.
        sources["x_mean"] = [v * units for v in sources["x_mean"]]
        sources["y_mean"] = [v * units for v in sources["y_mean"]]

    fwhm_x, fwhm_y = compute_fwhm(
        fake_image.image, sources, x_column="x_mean", y_column="y_mean"
    )

    expected_fwhm = np.array(sources["x_stddev"] * gaussian_sigma_to_fwhm)
    assert np.allclose(fwhm_x, expected_fwhm, rtol=1e-2)


@pytest.mark.parametrize("mask_by_nan", [True, False])
def test_compute_fwhm_with_missing_data(mask_by_nan):
    # Regression test for https://github.com/feder-observatory/stellarphot/issues/161
    # We should be able to find FWHM for a source even with NaNs in the image.
    fake_image = FakeImage(seed=SEED)
    sources = fake_image.sources
    x, y = sources["x_mean"].astype(int)[0], sources["y_mean"].astype(int)[0]
    image = fake_image.image.copy()

    if mask_by_nan:
        # Add a NaN to the image at the location of the first source. Note the
        # usual row/column swap when going to x/y coordinates.
        image[y, x] = np.nan
    else:
        image = CCDData(image, unit=u.adu, mask=np.zeros_like(image, dtype=bool))
        image.mask[y, x] = True

    # We expect a warning about NaNs in the image, so catch it
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=("Input data contains unmasked non-finite values "),
            category=AstropyUserWarning,
        )
    fwhm_x, fwhm_y = compute_fwhm(
        image, sources, x_column="x_mean", y_column="y_mean", fit_method=FwhmMethods.FIT
    )

    expected_fwhm = np.array(sources["x_stddev"] * gaussian_sigma_to_fwhm)
    assert np.allclose(fwhm_x, expected_fwhm, rtol=1e-2)


@pytest.mark.parametrize("image_is_ccd", [True, False])
@pytest.mark.parametrize("find_fwhm", [True, False])
@pytest.mark.parametrize("stddev_input", [True, False])
@pytest.mark.parametrize("spp_unit", [None, u.adu / u.pixel])
@pytest.mark.parametrize("provide_spp_input", [True, False, None])
def test_detect_source_number_location(
    provide_spp_input, spp_unit, stddev_input, find_fwhm, image_is_ccd
):
    """
    Make sure we detect the sources in the input table....
    """
    # Skip some cmobinations we really don't need to test
    if not provide_spp_input and spp_unit is not None:
        pytest.skip("No point in testing this combination")
    if image_is_ccd:
        fake_image = FakeCCDImage(seed=SEED)
        image = fake_image
    else:
        fake_image = FakeImage(seed=SEED)
        image = fake_image.image

    sources = QTable(
        fake_image.sources,
        units={
            "x_mean": u.pixel,
            "y_mean": u.pixel,
            "x_stddev": u.pixel,
            "y_stddev": u.pixel,
        },
    )
    # print(sources)
    if provide_spp_input:
        # Pass only one value for the sky background for source detection
        sky_per_pix = sources["sky_per_pix_avg"].mean()
        sky_per_pix *= spp_unit if spp_unit else 1
    else:
        sky_per_pix = None
        print(sources["sky_per_pix_avg"].mean())

    stddev = fake_image.noise_dev if stddev_input else None

    fwhm = 2 * sources["x_stddev"].mean()

    found_sources = source_detection(
        image,
        fwhm=fwhm,
        find_fwhm=find_fwhm,
        threshold=10,
        sky_per_pix_avg=sky_per_pix,
        stddev=stddev,
        verbose=True,  # Run verbose=True on this test to cover all verbose lines
    )
    # Sort by flux so we can reliably match them
    sources.sort("amplitude")
    found_sources.sort("flux")

    # Do we have the right number of sources?
    assert len(sources) == len(found_sources)

    for inp, out in zip(sources, found_sources, strict=True):
        # Do the positions match?
        np.testing.assert_allclose(out["xcenter"], inp["x_mean"], rtol=1e-5, atol=0.05)
        np.testing.assert_allclose(out["ycenter"], inp["y_mean"], rtol=1e-5, atol=0.05)
        if find_fwhm:
            np.testing.assert_allclose(
                gaussian_sigma_to_fwhm * (inp["x_stddev"] + inp["y_stddev"]) / 2,
                out["width"],
                rtol=1e-5,
                atol=0.05,
            )


def test_detect_source_with_padding():
    """
    Make sure we detect the sources in the input table....
    """
    fake_image = FakeImage(seed=SEED)
    sources = QTable(
        fake_image.sources,
        units={
            "x_mean": u.pixel,
            "y_mean": u.pixel,
            "x_stddev": u.pixel,
            "y_stddev": u.pixel,
        },
    )
    # Pass only one value for the sky background for source detection
    sky_per_pix = sources["sky_per_pix_avg"].mean()
    # Padding was chosen to be large enough to ensure that one of the sources in
    # test_sources.csv would land too close to the edge of the image.
    found_sources = source_detection(
        fake_image.image,
        fwhm=2 * sources["x_stddev"].mean(),
        threshold=10,
        sky_per_pix_avg=sky_per_pix,
        padding=95,
    )

    # Did we drop one source because it was too close to the edge?
    assert len(sources) - 1 == len(found_sources)


def test_detect_source_bad_input():
    with pytest.raises(ValueError, match="ccd must be a numpy array or CCDData object"):
        source_detection(None)


@pytest.mark.parametrize(
    "fit_method", [FwhmMethods.FIT, FwhmMethods.PROFILE, FwhmMethods.MOMENTS]
)
def test_fwhm_computation(fit_method):
    # Regression test for #490, in which FWHM computation is incorrect
    # because the image hasn't been background subtracted.
    ccd_file = get_pkg_data_path("data/cutout_for_fwhm_test.fits")
    source_list_file = get_pkg_data_path("data/source_list_for_fwhm_test.ecsv")

    ccd = CCDData.read(ccd_file, unit=u.adu)
    source_list = SourceListData.read(source_list_file)
    # Value below is from the image the cutout was taken from
    source_list["sky_per_pix_avg"] = np.array([83.69])

    fwhm_x, fwhm_y = compute_fwhm(
        ccd,
        source_list,
        fwhm_estimate=6.85179,
        x_column="xcenter",
        y_column="ycenter",
        sky_per_pix_column="sky_per_pix_avg",
        fit_method=fit_method,
    )

    avg_fwhm = np.mean([fwhm_x, fwhm_y])
    if fit_method == FwhmMethods.MOMENTS:
        assert avg_fwhm > 7
    else:
        assert np.isclose(avg_fwhm, 6.6, rtol=0.1)


def test_compute_fwhm_input_options():
    with pytest.raises(ValueError, match="Cannot specify both "):
        compute_fwhm(
            None, None, sky_per_pix_avg=10, sky_per_pix_column="sky_per_pix_avg"
        )

    # Make a table with a single column
    table = QTable({"sky_per_pix_avg": [10]})
    spp_column = "foo"
    with pytest.raises(
        ValueError, match=f"Column {spp_column} not found in sources table"
    ):
        compute_fwhm(None, table, sky_per_pix_column=spp_column)

    fake_image = FakeImage(seed=SEED)
    sources = fake_image.sources

    fit_method = "foo"
    with pytest.raises(ValueError, match=f"Unknown fit method: {fit_method}"):
        compute_fwhm(
            fake_image.image,
            sources,
            fit_method=fit_method,
            x_column="x_mean",
            y_column="y_mean",
        )


@pytest.mark.parametrize(
    "aggregate_by,n_bright_expect",
    (
        [None, 20],
        ["mean", 1],
        ["median", 1],
    ),
)
def test_fast_fwhm_from_image(aggregate_by, n_bright_expect):
    # This should maybe be split into a few tests. Putting that aside, this test
    # checks that
    # + the FWHM estimate is correct
    # + changing the max_adu changes the sources that are used
    expected_fwhm = 5.5
    fake_image = FakeImage(seed=SEED, fwhm=expected_fwhm, n_repeats_per_side=5)
    n_bright = 20
    fwhms = fast_fwhm_from_image(
        fake_image.image,
        5,
        noise=fake_image.noise_dev,
        n_brightest_sources=n_bright,
        max_adu=65000,
        aggregate_by=aggregate_by,
    )

    # Make sure we got the right number of sources
    assert len(np.atleast_1d(fwhms)) == n_bright_expect

    if aggregate_by is None:
        # Check that the FWHM values are reasonable
        assert np.allclose(fwhms, fake_image.input_fwhm, rtol=0.01)
    else:
        assert fwhms == pytest.approx(fake_image.input_fwhm, rel=0.01)

    # Make sure some of the stars are over the max_adu and check that
    # we get a different list of sources.
    max_adu = fake_image.sources["amplitude"].max() * 0.6
    fwhms_some_max_out = fast_fwhm_from_image(
        fake_image.image,
        5,
        noise=fake_image.noise_dev,
        n_brightest_sources=n_bright,
        max_adu=max_adu,
        aggregate_by=aggregate_by,
    )

    if aggregate_by is None:
        # Make sure we got different sources
        assert fwhms.mean() != fwhms_some_max_out.mean()
    else:
        assert fwhms != fwhms_some_max_out


def test_fast_fwhm_from_image_bad_aggregate():
    # check that providing a bad aggregate_by method raises an error
    expected_fwhm = 5.5
    fake_image = FakeImage(seed=SEED, fwhm=expected_fwhm, n_repeats_per_side=5)
    n_bright = 20
    with pytest.raises(ValueError, match="Unknown aggregate_by method"):
        fast_fwhm_from_image(
            fake_image.image,
            5,
            noise=fake_image.noise_dev,
            n_brightest_sources=n_bright,
            max_adu=65000,
            aggregate_by="foo",
        )



from pathlib import Path

import ipywidgets as ipw
from ipyautoui.custom import FileChooser

from stellarphot import PhotometryData
from stellarphot.settings.custom_widgets import Spinner

__all__ = ["TessAnalysisInputControls"]


class TessAnalysisInputControls(ipw.VBox):
    """
    A class to hold the widgets for choosing TESS input
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        hidden = ipw.Layout(display="none")

        self.phot_chooser = FileChooser(filter_pattern=["*.csv", "*.fits", "*.ecsv"])
        self._fits_openr = ipw.VBox(
            children=[
                ipw.HTML(value="<h3>Select your photometry/flux file</h3>"),
                self.phot_chooser,
            ]
        )
        self.tic_file_chooser = FileChooser(filter_pattern=["*.json"])
        fits_openr2 = ipw.VBox(
            children=[
                ipw.HTML(value="<h3>Select your TESS info file</h3>"),
                self.tic_file_chooser,
            ],
            layout=hidden,
        )
        self._passband = ipw.Dropdown(
            description="Ccoose Filter",
            options=["gp", "ip"],
            disabled=True,
            layout=hidden,
        )

        spinner = Spinner(message="<h4>Loading photometry...</h4>")

        self.phot_data = None

        def update_filter_list(_):
            spinner.start()
            self.phot_data = PhotometryData.read(self.phot_chooser.value)
            passband_data = self.phot_data["passband"]
            fits_openr2.layout.display = "flex"
            self._passband.layout.display = "flex"
            self._passband.options = sorted(set(passband_data))
            self._passband.disabled = False
            self._passband.value = self._passband.options[0]
            spinner.stop()

        self.phot_chooser.observe(update_filter_list, names="_value")
        self.children = [self._fits_openr, spinner, fits_openr2, self._passband]

    @property
    def tic_info_file(self):
        p = Path(self.tic_file_chooser.value)
        selected_file = p.name
        if not selected_file:
            raise ValueError("No TIC info json file selected")
        return p

    @property
    def photometry_data_file(self):
        p = Path(self.phot_chooser.value)
        selected_file = p.name
        if not selected_file:
            raise ValueError("No photometry data file selected")
        return p

    @property
    def passband(self):
        return self._passband.value


def filter_by_dates(
    phot_times=None,
    use_no_data_before=None,
    use_no_data_between=None,
    use_no_data_after=None,
):
    n_dropped = 0

    bad_data = phot_times < use_no_data_before

    n_dropped = bad_data.sum()

    if n_dropped > 0:
        print(
            f"👉👉👉👉 Dropping {n_dropped} data points before "
            f"BJD {use_no_data_before}"
        )

    bad_data = bad_data | (
        (use_no_data_between[0][0] < phot_times)
        & (phot_times < use_no_data_between[0][1])
    )

    new_dropped = bad_data.sum() - n_dropped

    if new_dropped:
        print(
            f"👉👉👉👉 Dropping {new_dropped} data points between "
            f"BJD {use_no_data_between[0][0]} and {use_no_data_between[0][1]}"
        )

    n_dropped += new_dropped

    bad_data = bad_data | (phot_times > use_no_data_after)

    new_dropped = bad_data.sum() - n_dropped

    if new_dropped:
        print(
            f"👉👉👉👉 Dropping {new_dropped} data points after "
            f"BJD {use_no_data_after}"
        )

    n_dropped += new_dropped
    return bad_data



import json

from astropy.io.misc.yaml import AstropyDumper, AstropyLoader

from stellarphot.settings import models

__all__ = []


def generate_table_representers(cls):
    """
    Call this method during initialization of a class to add the YAML
    Table representation for the class.
    """
    class_string = f"!{cls.__name__}"

    # Add YAML round-tripping for the model
    def _representer(dumper, model):
        # THIS SHOULD TAP INTO ASTROPY'S YAML DUMPER SOMEHOW
        # This is a little hacky at the moment. It seems like YAML
        # has trouble reading in a dictionary, so though model.model_dump()
        # works fine for writing, we can't construct from the dump.
        #
        # Instead of figuring out the right way to do that, this just dumps
        # the json representation as a string.
        return dumper.represent_mapping(
            class_string, {"model_json_string": model.model_dump_json()}
        )

    def _constructor(loader, node):
        # This loads the simple dictionary we dumped in _representer,
        # then initializes the model with the json string.
        mapping = loader.construct_mapping(node)
        return cls.model_validate_json(mapping["model_json_string"])

    AstropyDumper.add_representer(cls, _representer)
    AstropyLoader.add_constructor(class_string, _constructor)


def serialize_models_in_table_meta(table_meta):
    """
    Serialize the models in the table metadata **IN PLACE**.
    This is used to ensure that the models are represented as simple
    dictionaries when written to disk.

    Parameters
    ----------
    table_meta : dict
        The metadata dictionary of the table.
    """
    model_classes = tuple(getattr(models, model_name) for model_name in models.__all__)

    for key, value in table_meta.items():
        # If the value is a model instance, serialize it
        if isinstance(value, model_classes):
            model_instance = value
            # So, funny story. model_dump gives you a nice dictionary, in which
            # things like Longitude are turned into strings. However, writing them
            # to ECSV fails, because ECSV doesn't understand np.str_, and - guess what -
            # a Longitude returns a np.str_ when you do str(some_longitude).
            # The upshot is that the workaround here, i.e. using model_dump to get
            # simple objects into the header, does not work unless all string values
            # are converted to str.
            #
            # The issue has been reported in
            # https://github.com/astropy/astropy/issues/18235
            #

            # Dumping to json ensures that all the objects are converted to
            # very basic types, which is easy enough to convert to a dictionary.

            # Use model_dump_json to get a simple dictionary representation
            model_json = model_instance.model_dump_json()
            model_dict = json.loads(model_json)
            table_meta[key] = model_dict
            table_meta[key]["_model_name"] = model_instance.__class__.__name__
        # If the value is a dict, recurse
        elif isinstance(value, dict):
            serialize_models_in_table_meta(value)


def deserialize_models_in_table_meta(table_meta):
    """
    Deserialize the models in the table metadata **IN PLACE**.
    This is used to ensure that the models are restored from simple
    dictionaries when read from disk.

    There are two places a model might be stored:

    1. Directly in the table metadata.
    2. As a TableAttribute, which ends up in the table's meta under
       the "``__attributes__``" key.

    This function checks subdictionaries recursively to properly handle this
    case.

    Parameters
    ----------
    table_meta : dict
        The metadata dictionary of the table.
    """
    known_models = {
        model_name: getattr(models, model_name) for model_name in models.__all__
    }

    model_keys_in_meta = []
    for key, value in table_meta.items():
        # Check if the value is a dictionary and has a "_model_name" key
        if isinstance(value, dict):
            if "_model_name" in value:
                # Check if the model name is in the known models
                if value["_model_name"] in known_models.keys():
                    model_keys_in_meta.append(key)
            else:
                # Time to recurse into the dictionary
                deserialize_models_in_table_meta(value)

    for key in model_keys_in_meta:
        model_name = table_meta[key].pop("_model_name")
        table_meta[key] = known_models[model_name].model_validate(table_meta[key])


def _generate_old_table_representers():
    """
    This provides what is needed to read the "old-style" data tables in
    which the models were stored as objects in the table metadata.
    """
    for model_name in models.__all__:
        model_class = getattr(models, model_name)
        generate_table_representers(model_class)



import numpy as np
from astropy import units as u
from astropy.time import Time
from packaging.version import Version

from stellarphot import PhotometryData
from stellarphot.settings import Camera, Observatory, PassbandMap


class VersionMigrator:
    """
    Migrate data from one stellarphot major version to another.

    Parameters
    ----------
    from_version : str
        The version of the data to migrate from.

    to_version : str
        The version of the data to migrate to.

    camera : Camera
        The camera settings for the data.

    observatory : Observatory
        The observatory settings for the data.

    passband_map : PassbandMap
        The passband map for the data.
    """

    known_versions = ("1", "2")

    def __init__(
        self,
        from_version: str = "1",
        to_version: str = "2",
        camera: Camera = None,
        observatory: Observatory = None,
        passband_map: PassbandMap = None,
    ):
        self._from_version = Version(from_version)
        self._to_version = Version(to_version)
        if str(self.from_version.major) not in self.known_versions:
            raise ValueError(
                f"Unknown version: {self.from_version=}. "
                f"Valid versions are: {self.known_versions}"
            )

        if str(self.to_version.major) not in self.known_versions:
            raise ValueError(
                f"Unknown version: {self.to_version=}. "
                f"Valid versions are: {self.known_versions}"
            )

        if self.from_version >= self.to_version:
            raise ValueError(
                "Can only migrate from a lower version to a higher version"
            )

        self.camera = camera
        self.observatory = observatory
        self.passband_map = passband_map

    @property
    def from_version(self):
        """
        The stellarphot major version to migrate from.
        """
        return self._from_version

    @property
    def to_version(self):
        """
        The stellarphot major version to migrate to.
        """
        return self._to_version

    def migrate(self, data):
        """
        Migrate data from one version to another.

        Parameters
        ----------
        data : `astropy.table.Table` or `stellarphot.PhotometryData`
            The data to migrate.
        """
        if self.from_version.major == 1 and self.to_version.major == 2:
            return self._migrate_v1_v2(data)
        else:  # pragma: no cover
            # do not cover because right now no other version combo possible
            raise ValueError(
                f"Migration from version {self.from_version} to version "
                f"{self.to_version} is not supported."
            )

    def _migrate_v1_v2(self, data):
        """
        Migrate data from version 1 to version 2.

        Parameters
        ----------
        data : `astropy.table.Table` or `stellarphot.PhotometryData`
            The data to migrate.
        """
        new_data = data.copy()

        unitify = {
            "fwhm_x": u.pixel,
            "fwhm_y": u.pixel,
            "width": u.pixel,
            "aperture": u.pixel,
            "annulus_inner": u.pixel,
            "annulus_outer": u.pixel,
            "aperture_area": u.pixel,
            "annulus_area": u.pixel,
            "noise-aij": u.electron,
            "annulus_sum": u.adu,
            "aperture_sum": u.adu,
            "aperture_net_flux": u.adu,
            "noise": u.adu,
            "sky_per_pix_avg": u.adu / u.pixel,
            "sky_per_pix_med": u.adu / u.pixel,
            "sky_per_pix_std": u.adu / u.pixel,
            "RA": u.deg,
            "Dec": u.deg,
            "xcenter": u.pixel,
            "ycenter": u.pixel,
            "exposure": u.second,
        }
        for un in unitify:
            # Note that this deliberately discards any unit that might have been
            # present in the original data. This is a stellarphot-specific migration
            # tool, not something more general, and the choices above are the correct
            # units for the data.
            new_data[un] = new_data[un].value * unitify[un]

        new_data["date-obs"] = Time(new_data["date-obs"], scale="utc")

        v1_to_v2_col_renames = {
            "aperture_net_flux": "aperture_net_cnts",
            "BJD": "bjd",
            "noise": "noise_cnts",
            "noise-aij": "noise_electrons",
            "filter": "passband",
            "RA": "ra",
            "Dec": "dec",
        }

        # The new night calculation is better, so drop the old one
        del new_data["night"]

        # Handle changes in instrumental magnitude column names
        mag_columns = [col for col in new_data.colnames if col.startswith("mag_inst")]
        mag_cols_have_filter_names = any(
            mag.split("_")[-1] in new_data["filter"] for mag in mag_columns
        )

        if mag_cols_have_filter_names:
            new_mag_col_data = np.zeros_like(new_data["date-obs"])
            for col in mag_columns:
                passband = col.split("_")[-1]
                this_passband = new_data["filter"] == passband
                new_mag_col_data[this_passband] = new_data[col][this_passband]
            new_data["mag_inst"] = new_mag_col_data
            new_data.remove_columns(mag_columns)

        return PhotometryData(
            input_data=new_data,
            camera=self.camera,
            observatory=self.observatory,
            passband_map=self.passband_map,
            colname_map=v1_to_v2_col_renames,
        )



from copy import deepcopy

import pytest
from astropy import units as u
from astropy.table import Table
from packaging.version import InvalidVersion

from stellarphot import PhotometryData
from stellarphot.settings import Camera, Observatory
from stellarphot.settings.constants import TEST_CAMERA_VALUES, TEST_OBSERVATORY_SETTINGS
from stellarphot.utils.version_migrator import VersionMigrator

TEST_CAMERA_VALUES = deepcopy(TEST_CAMERA_VALUES)
TEST_OBSERVATORY_SETTINGS = deepcopy(TEST_OBSERVATORY_SETTINGS)


class TestVersionMigrator:
    @pytest.mark.parametrize(
        "from_version, to_version",
        [
            (None, None),
            ("1", None),
            (None, "2"),
            ("1", "2"),
        ],
    )
    def test_init(self, from_version, to_version):
        # Test a few valid ways to call the constructor
        vm_args = {}
        result_version = {}
        if from_version is not None:
            vm_args["from_version"] = from_version
            result_version["from_version"] = from_version
        else:
            result_version["from_version"] = "1"

        if to_version is not None:
            vm_args["to_version"] = to_version
            result_version["to_version"] = to_version
        else:
            result_version["to_version"] = "2"

        vm = VersionMigrator(**vm_args)
        assert str(vm.from_version) == result_version["from_version"]
        assert str(vm.to_version) == result_version["to_version"]

    def test_init_bad_versions(self):
        # Test some invalid version arguments

        # Migrator only increases version
        with pytest.raises(ValueError, match="Can only migrate from a lower version"):
            VersionMigrator(from_version="2", to_version="1")

        # Nothing to do if the versions are the same
        with pytest.raises(ValueError, match="Can only migrate from a lower version"):
            VersionMigrator(from_version="2", to_version="2")

        # An invalid version should raise an error
        with pytest.raises(InvalidVersion, match="not a parsable version"):
            VersionMigrator(from_version="not a parsable version")

        # An unknown from or to version should raise an error
        with pytest.raises(ValueError, match="Unknown version"):
            VersionMigrator(from_version="3")

        with pytest.raises(ValueError, match="Unknown version"):
            VersionMigrator(to_version="3")

    # The parameter two_filters is passed to the fixture stellphotv1_photometry_data
    # and is used to determine whether the test data contains data with one or
    # two filters.
    @pytest.mark.parametrize("two_filters", [True, False])
    @pytest.mark.parametrize("exposure_unit", [None, u.second])
    def test_migrate_v1_v2_mag_names_with_band(
        self, stellphotv1_photometry_data, exposure_unit
    ):
        # Test migrating from v1 to v2
        camera = Camera(**TEST_CAMERA_VALUES)
        observatory = Observatory(**TEST_OBSERVATORY_SETTINGS)
        vm = VersionMigrator(
            from_version="1", to_version="2", camera=camera, observatory=observatory
        )
        if exposure_unit is not None:
            stellphotv1_photometry_data["exposure"] *= exposure_unit
        sp2 = vm.migrate(stellphotv1_photometry_data)

        # Make sure the instrumental magnitudes have ended up in the right place
        for passband in set(sp2["passband"]):
            in_band_2 = sp2["passband"] == passband
            in_band_1 = stellphotv1_photometry_data["filter"] == passband
            assert all(
                sp2["mag_inst"][in_band_2]
                == stellphotv1_photometry_data[f"mag_inst_{passband}"][in_band_1]
            )

        # Make sure we can initialize a PhotometryData object from the migrated data
        # To ensure we go through the full init, change sp2 to a plain Table.
        input_data = Table(sp2)
        # Remove the autogenerated column
        del input_data["night"]

        # Make the new PhotometryData object
        sp2_pd = PhotometryData(input_data=input_data)
        # Check that the data is the same
        assert all(sp2_pd == sp2)
        # Check that we really got a new object
        assert sp2_pd is not sp2
        # Check the metadata is the same
        assert sp2_pd.meta == sp2.meta



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure.
import os

import pytest
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename
from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

from stellarphot import PhotometryData

# astropy-specific-stuff


def pytest_configure(config):
    from astropy.utils.iers import conf as iers_conf

    # Disable IERS auto download for testing
    iers_conf.auto_download = False

    config.option.astropy_header = True
    PYTEST_HEADER_MODULES.pop("h5py", None)

    # Always display versions of these packages
    for package in ["astroquery", "photutils", "astrowidgets", "pydantic", "ipyautoui"]:
        PYTEST_HEADER_MODULES[package] = package

    from .version import version

    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version


def pytest_unconfigure():
    from astropy.utils.iers import conf as iers_conf

    # Undo IERS auto download setting for testing
    iers_conf.reset("auto_download")


# stellarphot fixtures


@pytest.fixture
def profile_stars():
    # Make a few round stars
    return Table(
        dict(
            amplitude=[1000, 200, 300],
            x_mean=[30, 100, 150],
            y_mean=[40, 110, 160],
            x_stddev=[4, 4, 4],
            y_stddev=[4, 4, 4],
            theta=[0, 0, 0],
        )
    )


@pytest.fixture
def tess_tic_expected_values():
    # Use this in tests where you need a valid TOI (or TIC) and a coordinate for it.
    return dict(
        tic_id=236158940,
        expected_coords=SkyCoord(ra=313.41953739, dec=34.35164717, unit="degree"),
    )


@pytest.fixture
def simple_photometry_data():
    # Grab the test photometry file and simplify it a bit.
    data_file = get_pkg_data_filename("tests/data/test_photometry_data.ecsv")
    pd_input = PhotometryData.read(data_file)

    # Keep stars 1, 6, 9, 12, first time slice only
    # These stars have no NaNs in them and we only need one image to generate
    # more test data from that.
    first_slice = pd_input["file"] == "wasp-10-b-S001-R001-C099-r.fit"
    ids = [1, 6, 9, 12]
    good_star = pd_input["star_id"] == ids[0]

    for an_id in ids[1:]:
        good_star = good_star | (pd_input["star_id"] == an_id)

    return pd_input[good_star & first_slice]


@pytest.fixture
def stellphotv1_photometry_data(two_filters):
    """
    Load photometry data form version 1 of stellarphot.

    By default the data has two filters, with the filter name part of the column name,
    e.g. "mag_inst_B" and "mag_inst_ip".

    Parameters
    ----------
    two_filters : bool
        If True, return data with two filters. If False, return data with only the "B"
        filter and column name "mag_inst_B".
    """
    # Grab the test photometry file and simplify it a bit.
    data_file = get_pkg_data_filename("utils/tests/data/sp1-data-two-filters.csv")
    data = Table.read(data_file)
    if not two_filters:
        data = data[data["filter"] == "B"]
        del data["mag_inst_ip"]
    return data



import os
import warnings
from collections import namedtuple
from copy import deepcopy

import ipywidgets as ipw
import matplotlib
import pytest
from astropy.nddata import CCDData
from astrowidgets import ImageWidget
from photutils.datasets import make_noise_image

from stellarphot.gui_tools import (
    seeing_profile_functions as spf,
)
from stellarphot.gui_tools.seeing_profile_functions import (
    AP_SETTING_NEEDS_SAVE,
    AP_SETTING_SAVED,
)
from stellarphot.photometry.tests.fake_image import make_gaussian_sources_image
from stellarphot.photometry.tests.test_profiles import RANDOM_SEED, SHAPE
from stellarphot.settings import (
    Camera,
    Observatory,
    PhotometryApertures,
    PhotometryWorkingDirSettings,
    settings_files,  # This import is needed for mocking
)
from stellarphot.settings.constants import (
    TEST_CAMERA_VALUES,
)

TEST_CAMERA_VALUES = deepcopy(TEST_CAMERA_VALUES)


def test_keybindings():
    def simple_bindmap(bindmap):
        bound_keys = {}
        # The keys of the event map are...messy. This converts them to strings
        for key in bindmap.keys():
            modifier = key[1]
            key_name = key[2]
            bound_keys[str(key[0]) + "".join(modifier) + key_name] = key
        return bound_keys

    # This test assumes the ginga widget backend...
    iw = ImageWidget()
    original_bindings = iw._viewer.get_bindmap().eventmap

    bound_keys = simple_bindmap(original_bindings)
    # Spot check a couple of things before we run our function
    assert "Nonekp_D" in bound_keys
    assert "Nonekp_+" in bound_keys
    assert "Nonekp_left" not in bound_keys

    # rebind
    spf.set_keybindings(iw)
    new_bindings = iw._viewer.get_bindmap().eventmap
    bound_keys = simple_bindmap(new_bindings)
    assert "Nonekp_D" not in bound_keys
    assert "Nonekp_+" in bound_keys
    # Yes, the line below is correct...
    assert new_bindings[bound_keys["Nonekp_left"]]["name"] == "pan_right"


def test_seeing_profile_object_creation():
    # This test simply makes sure we can create the object
    profile_widget = spf.SeeingProfileWidget()
    assert isinstance(profile_widget.box, ipw.Box)


@pytest.fixture(autouse=True)
def fake_settings_dir(mocker, tmp_path):
    # See test_settings_files.py for more information on this fixture.
    # It makes a fake settings directory for each test to use.

    # stellarphot is added to the name of the directory to make sure we start
    # without a stellarphot directory for each test.
    mocker.patch.object(
        settings_files.PlatformDirs, "user_data_dir", tmp_path / "stellarphot"
    )


def test_seeing_profile_properties(tmp_path, profile_stars):
    # Here we make a seeing profile then load an image.
    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )

    # Make a fits file
    image = make_gaussian_sources_image(SHAPE, profile_stars) + make_noise_image(
        SHAPE, mean=10, stddev=100, seed=RANDOM_SEED
    )

    ccd = CCDData(image, unit="adu")
    ccd.header["exposure"] = 30.0
    ccd.header["object"] = "test"
    file_name = tmp_path / "test.fits"
    ccd.write(file_name)

    # Load the file
    profile_widget.fits_file.set_file("test.fits", tmp_path)
    profile_widget.load_fits()

    # Check that a couple of properties are set correctly
    assert profile_widget.object_name == "test"
    assert profile_widget.exposure == 30.0

    # Check that the photometry apertures have defaulted to the
    # default values in the model.
    assert profile_widget.aperture_settings.value == PhotometryApertures().model_dump()

    # Get the event handler that updates plots
    handler = profile_widget._make_show_event()

    # Make a mock event object
    Event = namedtuple("Event", ["data_x", "data_y"])
    star_loc_x, star_loc_y = profile_stars["x_mean"][0], profile_stars["y_mean"][0]
    # Sending a mock event will generate plots that we don't want to see
    # so set the matplotlib backend to a non-interactive one
    matplotlib.use("agg")
    # matplotlib generates a warning that we are using a non-interactive backend
    # so filter that warning out for the remainder of the test. There are at least
    # a couple of times we generate this warning as values are changed.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        handler(profile_widget.iw, Event(star_loc_x, star_loc_y))

        # The FWHM should be close to 9.6
        assert 9 < profile_widget.aperture_settings.value["fwhm_estimate"] < 10

        # Get a copy of the current aperture settings
        phot_aps = dict(profile_widget.aperture_settings.value)
        new_radius = phot_aps["radius"] - 2
        # Change the radius by directly setting the value of the widget that holds
        # the value. That ends up being nested fairly deeply...
        profile_widget.aperture_settings.di_widgets["radius"].value = new_radius

        # Make sure the settings are updated
        phot_aps["radius"] = new_radius
        assert profile_widget.aperture_settings.value == phot_aps


def test_seeing_profile_save_apertures(tmp_path):
    # Make sure that saving partial photometery settings works
    os.chdir(tmp_path)

    phot_settings = PhotometryWorkingDirSettings()

    # There should be no saved settings...
    with pytest.raises(ValueError, match="does not exist"):
        phot_settings.load()

    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )

    profile_widget.save()
    assert not profile_widget.aperture_settings.savebuttonbar.unsaved_changes

    settings = phot_settings.load()
    assert settings.camera == Camera(**TEST_CAMERA_VALUES)
    assert settings.photometry_apertures == PhotometryApertures(
        radius=1, annulus_width=1, gap=1
    )


def test_seeing_profile_save_box_title(tmp_path):
    # Save box title ends with different characters depending on whether values
    # need to be saved.

    # Change working directory since settings file is saved there.
    os.chdir(tmp_path)

    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )
    profile_widget.aperture_settings.di_widgets["radius"].value = 3

    assert profile_widget.aperture_settings.savebuttonbar.unsaved_changes
    assert AP_SETTING_NEEDS_SAVE in profile_widget.ap_title.value

    profile_widget.save()
    assert AP_SETTING_SAVED in profile_widget.ap_title.value


def test_seeing_profile_error_messages_no_star(tmp_path):
    # Make sure the appropriate error message is displayed when a click happens on
    # a region with no star, and that the message only appears once.
    profile_widget = spf.SeeingProfileWidget(
        camera=Camera(**TEST_CAMERA_VALUES), _testing_path=tmp_path
    )
    # Make a fits file with no stars
    image = make_noise_image(SHAPE, mean=10, stddev=1, seed=RANDOM_SEED)

    ccd = CCDData(image, unit="adu")
    ccd.header["exposure"] = 30.0
    ccd.header["object"] = "test"
    file_name = tmp_path / "test.fits"
    ccd.write(file_name)

    # Load the file
    profile_widget.fits_file.set_file("test.fits", tmp_path)
    profile_widget.load_fits()

    # Get the event handler that updates plots
    handler = profile_widget._make_show_event()

    # Make a mock event object
    Event = namedtuple("Event", ["data_x", "data_y"])
    star_loc_x, star_loc_y = SHAPE[0] // 2, SHAPE[1] // 2
    # Sending a mock event will generate plots that we don't want to see
    # so set the matplotlib backend to a non-interactive one
    matplotlib.use("agg")
    # matplotlib generates a warning that we are using a non-interactive backend
    # so filter that warning out for the remainder of the test. There are at least
    # a couple of times we generate this warning as values are changed.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert len(profile_widget.error_console.outputs) == 0

        # Clicking once should generate an error...
        handler(profile_widget.iw, Event(star_loc_x, star_loc_y))
        assert len(profile_widget.error_console.outputs) == 1
        assert (
            "No star found at this location"
            in profile_widget.error_console.outputs[0]["data"]["text/plain"]
        )

        # Clicking a second time should also just have one error
        handler(profile_widget.iw, Event(star_loc_x, star_loc_y))
        assert len(profile_widget.error_console.outputs) == 1


def test_seeing_profile_no_observatory():
    # This test checks that with no observatory set, there is no TESS
    # related box displayed.
    profile_widget = spf.SeeingProfileWidget()
    assert profile_widget.save_toggle is None
    assert profile_widget.tess_box.layout.visibility == "hidden"


@pytest.mark.parametrize("tess_code", [None, "dummy"])
def test_seeing_profile_with_observatory(tess_code):
    # Test two cases here:
    #  1. The observatory has TESS_telescope_code set to a string
    #  2. The observatory has TESS_telescope_code set to None
    observatory = Observatory(
        name="test",
        latitude=0.0,
        longitude=0.0,
        elevation="0.0 m",
        TESS_telescope_code=tess_code,
    )
    profile_widget = spf.SeeingProfileWidget(observatory=observatory)
    if tess_code is None:
        assert profile_widget.save_toggle is None
    else:
        assert isinstance(profile_widget.save_toggle, ipw.ToggleButton)



# Though these constants are used in the tests they are needed in a few
# non-test places too. Putting them here ensures they can be imported
# without needing test dependencies.

TEST_APERTURE_SETTINGS = dict(
    variable_aperture=False,
    radius=5,
    gap=10,
    annulus_width=15,
    fwhm_estimate=3.2,
)

TEST_CAMERA_VALUES = dict(
    data_unit="adu",
    gain="2.0 electron / adu",
    name="test camera",
    read_noise="10.0 electron",
    dark_current="0.01 electron / s",
    pixel_scale="0.563 arcsec / pix",
    max_data_value="50000.0 adu",
)

TEST_EXOPLANET_SETTINGS = dict(
    epoch={
        "jd1": 0.0,
        "jd2": 0.0,
        "format": "jd",
        "scale": "utc",
        "precision": 3,
        "in_subfmt": "*",
        "out_subfmt": "*",
    },
    period="0.0 min",
    identifier="a planet",
    coordinate={
        "ra": "0d00m00s",
        "dec": "0d00m00s",
        "representation_type": "spherical",
        "frame": "icrs",
    },
    depth=0,
    duration="0.0 min",
)

TEST_OBSERVATORY_SETTINGS = dict(
    name="test observatory",
    longitude="43d00m00s",
    latitude="45d00m00s",
    elevation="311.0 m",
    AAVSO_code="test",
    TESS_telescope_code="tess test",
)

# The first setting here is required, the rest are optional. The optional
# settings below are different than the defaults in the model definition.
TEST_PHOTOMETRY_OPTIONS = dict(
    include_dig_noise=False,
    reject_too_close=False,
    reject_background_outliers=False,
    fwhm_method="fit",
    partial_pixel_method="center",
)

TEST_PASSBAND_MAP = dict(
    name="Example map",
    your_filter_names_to_aavso=[
        dict(
            your_filter_name="V",
            aavso_filter_name="V",
        ),
        dict(
            your_filter_name="B",
            aavso_filter_name="B",
        ),
        dict(
            your_filter_name="rp",
            aavso_filter_name="SR",
        ),
    ],
)

TEST_LOGGING_SETTINGS = dict(
    logfile="test.log",
    console_log=False,
)

TEST_SOURCE_LOCATION_SETTINGS = dict(
    shift_tolerance=5,
    source_list_file="test.ecsv",
    use_coordinates="pixel",
)

TEST_PHOTOMETRY_SETTINGS = dict(
    camera=TEST_CAMERA_VALUES,
    observatory=TEST_OBSERVATORY_SETTINGS,
    photometry_apertures=TEST_APERTURE_SETTINGS,
    source_location_settings=TEST_SOURCE_LOCATION_SETTINGS,
    photometry_optional_settings=TEST_PHOTOMETRY_OPTIONS,
    passband_map=TEST_PASSBAND_MAP,
    logging_settings=TEST_LOGGING_SETTINGS,
)
